import numpy as np

from QuICT.core.gate import CompositeGate, CX, CCX, CSwap, X, QFT, IQFT, CU1, U1, CCRz, Phase


def ex_gcd(a, b, arr):
    if b == 0:
        arr[0] = 1
        arr[1] = 0
        return a
    g = ex_gcd(b, a % b, arr)
    t = arr[0]
    arr[0] = arr[1]
    arr[1] = t - int(a / b) * arr[1]
    return g


def mod_reverse(a, n):
    arr = [0, 1]
    g = ex_gcd(a, n, arr)
    if g != 1:
        raise ValueError("not coprime")
    return (arr[0] % n + n) % n


def draper_adder(gate_set, a, b):
    """ store a + b in b

    (a,b) -> (a,b'=a+b)


    Args:
        gate_set(CompositeGate):
        a(list): the list of indices of qureg stores a, length is n
        b(list): the list of indices of qureg stores b, length is n

    """
    n = len(a)
    with gate_set:
        QFT.build_gate(n) & b
        for i in range(n):
            p = 0
            for j in range(i, n):
                p += 1
                CU1(2 * np.pi / (1 << p)) & [a[j], b[i]]
        IQFT.build_gate(n) & b


def fourier_adder_wired(gate_set, a, phib):
    """ store Φ(a + b) in phib, but a is wired

    (phib) -> (phib'=Φ(a + b))

    Args:
        a(int):      unsigned integer, low n bits used
        phib(Qureg): the qureg stores Φ(b), length is n+1

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """
    n = len(phib) - 1
    a = a & ~(1 << n)  # clear (n+1)-th bit to zero
    with gate_set:
        for i in range(n + 1):
            p = 0
            for j in range(i, n + 1):
                p += 1
                if a & (1 << (n - j)) != 0:
                    U1(2 * np.pi / (1 << p)) & phib[i]


def fourier_adder_wired_reversed(gate_set, a, phib):
    """ store Φ(b - a) or Φ(b - a + 2**(n+1)) in phib, but a is wired

    (phib) -> (phib')

    Args:
        a(int):      unsigned integer, low n bits used
        phib(Qureg): the qureg stores Φ(b), length is n+1

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """
    n = len(phib) - 1
    a = a & ~(1 << n)  # clear (n+1)-th bit to zero
    with gate_set:
        for i in reversed(range(n + 1)):
            p = n + 1 - i
            for j in reversed(range(i, n + 1)):
                if a & (1 << (n - j)) != 0:
                    U1(- 2 * np.pi / (1 << p)) & phib[i]
                p -= 1


def cc_fourier_adder_wired(gate_set, a, phib, c, dualControlled):
    """ fourier_adder_wired with 1 or 2 control bits

    (phib,c) -> (phib'=Φ(a + b),c)

    Args:
        a(int):      unsigned integer, low n bits used
        phib(Qureg): the qureg stores Φ(b), length is n
        c(Qureg):    the control qubits,    length is 2 or 1, see dualControlled
        dualControlled(bool): if True, c[0] will be used; else c[0:2] will be used

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """
    if type(c) == int:
        c = [c]
    n = len(phib) - 1
    a = a & ~(1 << n)  # clear (n+1)-th bit to zero
    with gate_set:
        for i in range(n + 1):
            p = 0
            for j in range(i, n + 1):
                p += 1
                if a & (1 << (n - j)) != 0:
                    phase = 2 * np.pi / (1 << p)
                    # CCU1(2 * np.pi / (1 << p)) | (c[0],c[1],phib[i])
                    if dualControlled:
                        CU1(phase / 2) & [c[1], phib[i]]
                        CX & [c[0], c[1]]
                        CU1(- phase / 2) & [c[1], phib[i]]
                        CX & [c[0], c[1]]
                        CU1(phase / 2) & [c[0], phib[i]]
                    else:
                        CU1(phase) & [c[0], phib[i]]


def cc_fourier_adder_wired_reversed(gate_set, a, phib, c, dualControlled):
    """ fourier_adder_wired_reversed with 1 or 2 control bits

    (phib,c) -> (phib',c)

    Args:
        a(int):      unsigned integer, low n bits used
        phib(Qureg): the qureg stores Φ(b), length is n+1
        c(Qureg):    the control qubits,    length is 2 or 1, see dualControlled
        dualControlled(bool): default True. if True, c[0] will be used; else c[0:1] will be used

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """
    if type(c) == int:
        c = [c]
    n = len(phib) - 1
    a = a & ~(1 << n)  # clear (n+1)-th bit to zero
    with gate_set:
        for i in reversed(range(n + 1)):
            p = n + 1 - i
            for j in reversed(range(i, n + 1)):
                if a & (1 << (n - j)) != 0:
                    phase = 2 * np.pi / (1 << p)
                    # CCU1(2 * np.pi / (1 << p)) | (c[0],c[1],phib[i])
                    if dualControlled:
                        CU1(- phase / 2) & [c[0], phib[i]]
                        CX & [c[0], c[1]]
                        CU1(phase / 2) & [c[1], phib[i]]
                        CX & [c[0], c[1]]
                        CU1(- phase / 2) & [c[1], phib[i]]
                    else:
                        CU1(- phase) & [c[0], phib[i]]
                p -= 1


def cc_fourier_adder_mod(gate_set, a, N, phib, c, low, dualControlled=True):
    """ use fourier_adder_wired/cc_fourier_adder_wired to calculate (a+b)%N in Fourier space

    (phib=Φ(b),c,low) -> (phib'=Φ((a+b)%N),c,low)


    Args:
        a(int):      low n bits used as unsigned
        N(int):      low n bits used as unsigned
        phib(Qureg): the qureg stores b,        length is n+1,
        c(Qureg):    the control qubits,    length is 2 or 1, see dualControlled
        low(Qureg):  the clean ancillary qubit, length is 1,
        dualControlled(bool): if True, c[0] will be used; else c[0:1] will be used

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """
    cc_fourier_adder_wired(gate_set, a, phib, c, dualControlled=dualControlled)
    fourier_adder_wired_reversed(gate_set, N, phib)
    with gate_set:
        IQFT.build_gate(len(phib)) & phib
        CX & [phib[0], low[0]]
        QFT.build_gate(len(phib)) & phib
    cc_fourier_adder_wired(gate_set, N, phib, low, dualControlled=False)
    cc_fourier_adder_wired_reversed(
        gate_set, a, phib, c, dualControlled=dualControlled)
    with gate_set:
        IQFT.build_gate(len(phib)) & phib
        X & phib[0]
        CX & [phib[0], low[0]]
        X & phib[0]
        QFT.build_gate(len(phib)) & phib
    cc_fourier_adder_wired(gate_set, a, phib, c, dualControlled=dualControlled)


def fourier_adder_mod(gate_set, a, N, phib, low):
    """ use fourier_adder_wired/cc_fourier_adder_wired
    to calculate (a+b)%N in Fourier space. no control bits.

    (phib=Φ(b),low) -> (phib'=Φ((a+b)%N),low)

    Args:
        a(int):      low n bits used as unsigned
        N(int):      low n bits used as unsigned
        phib(Qureg): the qureg stores b,        length is n+1,
        low(Qureg):  the clean ancillary qubit, length is 1,

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """
    fourier_adder_wired(gate_set, a, phib)
    fourier_adder_wired_reversed(gate_set, N, phib)
    with gate_set:
        IQFT.build_gate(len(phib)) & phib
        CX & [phib[0], low[0]]
        QFT.build_gate(len(phib)) & phib
    cc_fourier_adder_wired(gate_set, N, phib, low, dualControlled=False)
    fourier_adder_wired_reversed(gate_set, a, phib)
    with gate_set:
        IQFT.build_gate(len(phib)) & phib
        X & phib[0]
        CX & [phib[0], low[0]]
        X & phib[0]
        QFT.build_gate(len(phib)) & phib
    fourier_adder_wired(gate_set, a, phib)


def c_fourier_mult_mod(gate_set, a, N, x, phib, c, low):
    """ use cc_fourier_adder_mod to calculate (b+ax)%N in Fourier space

    (phib=Φ(b),x,c,low) -> (phib'=Φ((b+ax)%N),x,c,low)

    Args:
        a(int):      low n bits used as unsigned
        N(int):      low n bits used as unsigned
        x(Qureg):    the qureg stores x,        length is n,
        phib(Qureg): the qureg stores b,        length is n+1,
        c(int):      the control qubits,        length is 1,
        low(int):    the clean ancillary qubit, length is 1,

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """

    n = len(phib) - 1
    p = 1
    for i in range(n - 1, -1, -1):
        cc_fourier_adder_mod(gate_set, p * a % N, N, phib,
                             (c[0], x[i]), low)  # p * a % N
        p = p * 2


def fourier_mult_mod(gate_set, a, N, x, phib, low):
    """ use fourier_adder_mod to calculate (b+ax)%N in Fourier space

    (phib=Φ(b),x,low) -> (phib'=Φ((b+ax)%N),x,low)

    Args:
        a(int):      low n bits used as unsigned
        N(int):      low n bits used as unsigned
        x(Qureg):    the qureg stores x,        length is n,
        phib(Qureg): the qureg stores b,        length is n+1,
        low(Qureg):  the clean ancillary qubit, length is 1,

    Circuit for Shor's algorithm using 2n+3 qubits
    http://arxiv.org/abs/quant-ph/0205095v3
    """

    n = len(phib) - 1
    p = 1
    for i in range(n - 1, -1, -1):
        cc_fourier_adder_mod(gate_set, p * a % N, N, phib,
                             x[i], low, dualControlled=False)  # p * a % N
        p = p * 2


def c_mult_mod(gate_set, a, N, x, b, c, low):
    with gate_set:
        QFT.build_gate(len(b)) & b
        c_fourier_mult_mod(gate_set, a, N, x, b, c, low)
        IQFT.build_gate(len(b)) & b


class BEAAdder(object):
    @staticmethod
    def execute(n):
        """ a circuit calculate a+b, a and b are gotten from some qubits.

        (a,b) -> (a,b'=a+b)

        Args:
            n(int): length of a and b
        """

        gate_set = CompositeGate()
        qreg_a = list(range(n))
        qreg_b = list(range(n, n * 2))
        draper_adder(gate_set, qreg_a, qreg_b)
        return gate_set


class BEAAdderWired(object):
    @staticmethod
    def execute(n, a):
        """ a circuit calculate a+b, a is wired, and b are gotten from some qubits.

        (b) -> (b'=a+b)

        Args:
            n(int): length of a. b is in length n+1
            a(int): the operand to be added. low n bits used
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            fourier_adder_wired(gate_set, a, qreg_b)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class BEAReverseAdderWired(object):
    @staticmethod
    def execute(n, a):
        """
        (b) -> (b'=b-a or b-a+2**(n+1))

        Args:
            n(int): length of a. b is in length n+1
            a(int): the operand to be subtracted. low n bits used
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            fourier_adder_wired_reversed(gate_set, a, qreg_b)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class CCBEAAdderWired(object):
    @staticmethod
    def execute(n, a):
        """
        (b,c) -> (b'=a+b,c) if c=0b11 else (b'=b,c)

        Args:
            n(int): length of a. b is in length n+1
            a(int): the operand to be subtracted. low n bits used
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_c = list(range(n + 1, n + 3))
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            cc_fourier_adder_wired(gate_set, a, qreg_b,
                                   qreg_c, dualControlled=True)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class CCBEAReverseAdderWired(object):
    @staticmethod
    def execute(n, a):
        """
        (b,c) -> (b'=b-a,c) if c=0b11 else (b'=b,c)

        Args:
            n(int): length of a. b is in length n+1
            a(int): the operand to be subtracted. low n bits used
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_c = list(range(n + 1, n + 3))
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            cc_fourier_adder_wired_reversed(
                gate_set, a, qreg_b, qreg_c, dualControlled=True)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class CCBEAAdderMod(object):
    @staticmethod
    def execute(n, a, N):
        """ use fourier_adder_wired/cc_fourier_adder_wired to calculate (a+b)%N in Fourier space

        (phib=Φ(b),c,low) -> (phib'=Φ((a+b)%N),c,low) if c=0b11 else (phib'==Φ(b),c,low)

        Args:
            n(int):      bits len
            a(int):      least n bits used as unsigned
            N(int):      least n bits used as unsigned

        Quregs:
            phib(Qureg): the qureg stores b,        length is n+1,
            c(Qureg):    the control qubits,        length is 2,
            low(Qureg):  the clean ancillary qubit, length is 1,

        Circuit for Shor's algorithm using 2n+3 qubits
        http://arxiv.org/abs/quant-ph/0205095v3
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_c = list(range(n + 1, n + 3))
        qreg_low = list(range(n + 3, n + 4))
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            cc_fourier_adder_mod(gate_set, a, N, qreg_b, qreg_c, qreg_low)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class BEAAdderMod(object):
    @staticmethod
    def execute(n, a, N):
        """ use fourier_adder_wired/cc_fourier_adder_wired
        to calculate (a+b)%N in Fourier space. No cotrol bits

        (phib=Φ(b),low) -> (phib'=Φ((a+b)%N),low)

        Args:
            n(int):      bits len
            a(int):      least n bits used as unsigned
            N(int):      least n bits used as unsigned

        Quregs:
            phib(Qureg): the qureg stores b,        length is n+1,
            low(Qureg):  the clean ancillary qubit, length is 1,

        Circuit for Shor's algorithm using 2n+3 qubits
        http://arxiv.org/abs/quant-ph/0205095v3
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_low = list(range(n + 1, n + 2))
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            fourier_adder_mod(gate_set, a, N, qreg_b, qreg_low)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class CBEAMulMod(object):
    @staticmethod
    def execute(n, a, N):
        """ use cc_fourier_adder_mod to calculate (b+ax)%N in Fourier space

        (phib=Φ(b),x,c,low) -> (phib'=Φ((b+ax)%N),x,c,low) if c=0b1 else (phib'=Φ(b),x,c,low)

        Args:
            n(int):      bits len
            a(int):      least n bits used as unsigned
            N(int):      least n bits used as unsigned

        Quregs:
            phib(Qureg): the qureg stores b,        length is n+1,
            x(Qureg):    the qureg stores x,        length is n,
            c(Qureg):    the control qubits,        length is 1,
            low(Qureg):  the clean ancillary qubit, length is 1,

        Circuit for Shor's algorithm using 2n+3 qubits
        http://arxiv.org/abs/quant-ph/0205095v3
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_x = list(range(n + 1, 2 * n + 1))
        qreg_c = [2 * n + 1]
        qreg_low = [2 * n + 2]
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            c_fourier_mult_mod(gate_set, a, N, qreg_x,
                               qreg_b, qreg_c, qreg_low)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class BEAMulMod(object):
    @staticmethod
    def execute(n, a, N):
        """ use fourier_adder_mod to calculate (b+ax)%N in Fourier space. No control bits

        (phib=Φ(b),x,low) -> (phib'=Φ((b+ax)%N),x,low)


        Args:
            n(int):      bits len
            a(int):      least n bits used as unsigned
            N(int):      least n bits used as unsigned

        Quregs:
            phib(Qureg): the qureg stores b,        length is n+1,
            x(Qureg):    the qureg stores x,        length is n,
            low(Qureg):  the clean ancillary qubit, length is 1,

        Circuit for Shor's algorithm using 2n+3 qubits
        http://arxiv.org/abs/quant-ph/0205095v3
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_x = list(range(n + 1, 2 * n + 1))
        qreg_low = [2 * n + 1]
        with gate_set:
            QFT.build_gate(len(qreg_b)) & qreg_b
            fourier_mult_mod(gate_set, a, N, qreg_x, qreg_b, qreg_low)
            IQFT.build_gate(len(qreg_b)) & qreg_b
        return gate_set


class BEACUa(object):
    @staticmethod
    def execute(n, a, N):
        """ Controlled-U_a, ((a*x)MOD(N)) if c=1, else (x)

        (b=0,x,c,low) -> (b=0,x',c,low)

        Args:
            n(int):      bits len
            a(int):      least n bits used as unsigned
            N(int):      least n bits used as unsigned

        Quregs:
            phib(Qureg): the qureg stores b,        length is n+1,
            x(Qureg):    the qureg stores x,        length is n,
            c(Qureg):    the qureg stores c,        length is 1,
            low(Qureg):  the clean ancillary qubit, length is 1,

        Circuit for Shor's algorithm using 2n+3 qubits
        http://arxiv.org/abs/quant-ph/0205095v3
        """

        gate_set = CompositeGate()
        qreg_b = list(range(n + 1))
        qreg_x = list(range(n + 1, 2 * n + 1))
        qreg_c = [2 * n + 1]
        qreg_low = [2 * n + 2]

        with gate_set:
            gate_set: CompositeGate
            c_mult_mod(gate_set, a, N, qreg_x, qreg_b, qreg_c, qreg_low)
            for i in range(n):  # n bits swapped, b[0] always 0
                CSwap & [qreg_c[0], qreg_x[i], qreg_b[i + 1]]
            # Reverse c_mult_mod(a_inv,N,x,b,c,low)
            c_mult_mod(gate_set, N - mod_reverse(a, N), N, qreg_x, qreg_b, qreg_c, qreg_low)

        return gate_set



#!/usr/bin/env python
# -*- coding:utf8 -*-

import heapq
from typing import *
from .graph import *
from .utility import *


class EdgeColoring:
    """
    This class provides a quick algorithm to decompose
    a bipartite graph into multiple matching.
    """

    def __init__(self):
        pass

    @staticmethod
    def _merge_one_side(bipartite: Bipartite, is_left: bool, mx_deg: int) -> List[Merged]:
        """

        Args:
            bipartite(Bipartite) : exec merge on this bipartite
            is_left(bool) : Whether it's left
            mx_deg(int) : max degree of bipartite


        Returns:
            List[Merged] : Merged side
        """
        side = copy.copy(bipartite.left) if is_left else copy.copy(bipartite.right)
        hq = []
        for i in side:
            heapq.heappush(hq, Merged(
                bipartite.get_degree(i),
                [i]
            ))
        while len(hq) >= 2:
            a: Merged = heapq.heappop(hq)
            b: Merged = heapq.heappop(hq)
            if a.deg + b.deg < mx_deg:
                heapq.heappush(hq, a + b)
            else:
                heapq.heappush(hq, a)
                heapq.heappush(hq, b)
                break
        return hq

    @classmethod
    def get_regular(cls, bipartite: Bipartite) -> Tuple[dict, dict, Bipartite]:
        """
        Get a regular graph(all nodes have the same degree) from a bipartite graph.
        After the transformation, some indices of nodes might be merged.

        Args:
            bipartite(Bipartite) : a bipartite graph

        Returns:
            A tuple of (new_to_olds, old_to_new, regular_graph). Where new_to_olds
            is a dict from new index to old indices, old_to_new as the one from the
            old to the new and regular_graph is the graph we get.
        """
        new_to_olds = {}
        """int-int dict
        """
        old_to_new = {}
        """int-List[int] dict
        """
        cnt = 1
        left = []
        right = []

        # merge left side
        mx_deg = bipartite.get_max_degree()
        hq = cls._merge_one_side(bipartite, True, mx_deg)
        for old in hq:
            new_to_olds[cnt] = old.nodes
            left.append(cnt)
            cnt += 1

        # merge right side
        hq: List[Merged] = cls._merge_one_side(bipartite, False, mx_deg)
        for old in hq:
            new_to_olds[cnt] = old.nodes
            right.append(cnt)
            cnt += 1

        while len(left) < len(right):
            new_to_olds[cnt] = []
            left.append(cnt)
            cnt += 1
        while len(right) < len(left):
            new_to_olds[cnt] = []
            right.append(cnt)
            cnt += 1

        for new_idx in new_to_olds:
            for idx in new_to_olds[new_idx]:
                old_to_new[idx] = new_idx

        regular_graph = Bipartite(left, right)
        # for edge in bipartite.edges:
        #     s = old_to_new[edge.start]
        #     e = old_to_new[edge.end]
        #     regular_graph.add_edge(s, e, mark=1)

        # you have to add it pairwise
        for node in bipartite.left:
            eid = bipartite.head[node]
            while eid != -1:
                edge = bipartite.edges[eid]
                s = old_to_new[edge.start]
                e = old_to_new[edge.end]
                regular_graph.add_edge(s, e, valid=True)
                regular_graph.add_edge(e, s, valid=True)
                eid = edge.next

        # add some extra "fake" edges
        hql = []
        hqr = []
        heapq.heapify(hql)
        heapq.heapify(hqr)
        for i in regular_graph.left:
            heapq.heappush(hql, (regular_graph.get_degree(i), i))
        for i in regular_graph.right:
            heapq.heappush(hqr, (regular_graph.get_degree(i), i))
        while True:
            a = heapq.heappop(hql)
            b = heapq.heappop(hqr)
            if a[0] == mx_deg:
                break
            regular_graph.add_edge(a[1], b[1], valid=False)
            regular_graph.add_edge(b[1], a[1], valid=False)
            heapq.heappush(hql, (a[0] + 1, a[1]))
            heapq.heappush(hqr, (b[0] + 1, b[1]))
        return new_to_olds, old_to_new, regular_graph

    @classmethod
    def hungarian_dfs(
            cls,
            color: int,
            bipartite: Bipartite,
            cur: int,
            in_alter_path: List[bool],
            matching: List[int]
    ) -> bool:
        """
        Dfs used in hungarian algorithm

        Args:
            color(int) : mark this color for matching edges
            bipartite(Bipartite) : execute dfs on this bipartite
            cur(int) : start point of current dfs
            in_alter_path(List[bool]) : if some point is in an alternative path
            matching(List[int]) : match point, -1 for no match

        Returns:
            bool : dfs success
        """

        eid: int = bipartite.head[cur]  # edge index
        while eid != -1:
            edge = bipartite.edges[eid]
            to = edge.end
            eid = edge.next
            if not in_alter_path[to]:
                in_alter_path[to] = True
                if matching[to] == -1 or cls.hungarian_dfs(
                        color,
                        bipartite,
                        matching[to],
                        in_alter_path,
                        matching
                ):  # :(
                    matching[to] = cur
                    matching[cur] = to
                    return True
        return False

    @classmethod
    def get_colored_matching(
            cls,
            bipartite: Bipartite,
            color: int,
            change_input: bool = True
    ) -> Tuple[int, Bipartite, Bipartite]:
        """
        Get a matching of a regular bipartite graph and color it if needed.
        Remember that a regular graph always has a perfect matching.

        In fact, in the original paper, matching is found using a O(ElogD) algorithm.
        It's too difficult to implement. Use Hungarian instead. Time complexity is increased
        but it has no effect on the ultimate cycle.

        TODO: replace Hungarian matching algorithm(which runs in O(VE)) for a O(ElogE) or O(ElogD) one.

        Args:
            bipartite(Bipartite) : Exec colored matching on this bipartite
            color(int) : Integer index of some color
            change_input(bool) : Whether put color on the original graph

        Returns:
            Tuple[int, Bipartite, Bipartite]: Matching edge number, Match bipartite, Bipartite removed matching.
        """
        n = len(bipartite.nodes)
        matching = [-1 for _ in range(n + 1)]
        cnt = 0
        for node in bipartite.left:
            if matching[node] == -1:
                in_alter_path = [False for _ in range(n + 1)]
                # DO NOT change color during dfs cause there is possible backtracing.
                if cls.hungarian_dfs(color, bipartite, node, in_alter_path, matching):
                    cnt += 1

        matching_bipartite = Bipartite(bipartite.left, bipartite.right)
        for node in matching_bipartite.left:
            other = matching[node]
            if other == -1:
                continue
            s = node
            e = other
            matching_bipartite.add_edge(start=s, end=e, valid=True, color=color)
            matching_bipartite.add_edge(start=e, end=s, valid=True, color=color)

            # Color original bipartite if needed.
            if change_input:
                eid = bipartite.head[s]
                while eid != -1:
                    edge = bipartite.edges[eid]
                    if edge.end == e:
                        edge.color = color
                        break
                    eid = edge.next
                eid = bipartite.head[e]
                while eid != -1:
                    edge = bipartite.edges[eid]
                    if edge.end == s:
                        edge.color = color
                        break
                    eid = edge.next

        extracted_bipartite = Bipartite(bipartite.left, bipartite.right)
        for node in bipartite.left:
            eid = bipartite.head[node]
            while eid != -1:
                edge = bipartite.edges[eid]
                s = edge.start
                e = edge.end
                if matching[s] == e:
                    matching[s] = -1
                    matching[e] = -1
                    eid = edge.next
                    continue
                extracted_bipartite.add_edge(s, e)
                extracted_bipartite.add_edge(e, s)
                eid = edge.next
        assert len(extracted_bipartite.edges) + len(matching_bipartite.edges) == len(bipartite.edges)
        return cnt, matching_bipartite, extracted_bipartite

    @classmethod
    def get_euler_partition(cls, bipartite: Bipartite, skip_colored: bool = True) -> Tuple[Bipartite, Bipartite]:
        """
        Split original 2r-regular bipartite into 2 r-regular bipartite.


        Args:
            bipartite(Bipartite) : Split this.
            skip_colored(bool) : Whether to skip the colored edges in the graph.
            If true, edges without color should form a 2r-regular graph.

        Returns:
            Tuple[Bipartite, Bipartite] : 2 bipartite after split
        """
        e = len(bipartite.edges)
        b1 = Bipartite(bipartite.left, bipartite.right)
        b2 = Bipartite(bipartite.left, bipartite.right)
        """
        Original graph might not be connected, se we have to iterate over
        all possible connected component.
        """
        visited = [False for _ in range(e + 1)]
        nid = 0
        n_cnt = len(bipartite.left)
        visited_cnt = 0
        left_turn = True
        while nid < n_cnt:
            cur = bipartite.left[nid]
            while True:
                """
                Actually, for 2r-regular bipartite graph, that has_out==False
                is only possible when we exhaust a connected component and return
                to start point.
                """
                # check if there's an outward edge
                eid = bipartite.head[cur]
                while eid != -1:
                    edge = bipartite.edges[eid]
                    # skip visited edges
                    if visited[eid]:
                        eid = edge.next
                        continue
                    # skip colored edges
                    elif skip_colored and edge.color != -1:
                        eid = edge.next
                        continue
                    break
                if eid == -1:
                    break
                visited[eid] = True
                visited[eid ^ 1] = True
                visited_cnt += 1
                end = bipartite.edges[eid].end
                if left_turn:
                    b1.add_edge(cur, end)
                    b1.add_edge(end, cur)
                else:
                    b2.add_edge(cur, end)
                    b2.add_edge(end, cur)
                left_turn = not left_turn
                cur = end
            nid += 1
        # assert e // 2 == visited_cnt
        # print(f"undirected edges: {e // 2}, visited: {visited_cnt}")
        return b1, b2

    @classmethod
    def _cpy_edge(cls, _b: Bipartite, _ret: Bipartite):
        for _node in _b.left:
            _eid = _b.head[_node]
            while _eid != -1:
                _edge = _b.edges[_eid]
                _ret.add_edge(
                    start=_edge.start,
                    end=_edge.end,
                    valid=_edge.valid,
                    color=_edge.color
                )
                _ret.add_edge(
                    start=_edge.end,
                    end=_edge.start,
                    valid=_edge.valid,
                    color=_edge.color
                )
                _eid = _edge.next

    @classmethod
    def get_edge_coloring_for_regular(cls, bipartite: Bipartite, start_color: int) -> Bipartite:
        """
        Coloring this regular bipartite

        Args:
            bipartite(Bipartite) : exec on this bipartite
            start_color(int) : color used would start with this index

        Returns:
            Bipartite
        """

        ret = Bipartite(bipartite.left, bipartite.right)
        deg = bipartite.get_any_degree()
        # assert deg > 0
        if deg == 0:  # Sometimes there is really a empty graph
            return ret
        if deg == 1:
            for node in bipartite.left:
                eid = bipartite.head[node]
                if eid == -1:
                    continue
                edge = bipartite.edges[eid]
                ret.add_edge(
                    start=edge.start,
                    end=edge.end,
                    valid=edge.valid,
                    color=start_color
                )
                ret.add_edge(
                    start=edge.end,
                    end=edge.start,
                    valid=edge.valid,
                    color=start_color
                )
        elif deg % 2 == 0:
            b1, b2 = cls.get_euler_partition(bipartite)
            b3 = cls.get_edge_coloring_for_regular(b1, start_color)
            b4 = cls.get_edge_coloring_for_regular(b2, start_color + deg // 2)

            cls._cpy_edge(b3, ret)
            cls._cpy_edge(b4, ret)
        else:  # deg%2 == 1
            _, b1, b2 = cls.get_colored_matching(bipartite, start_color)
            cls._cpy_edge(b1, ret)
            start_color += 1
            b3, b4 = cls.get_euler_partition(b2)
            b5 = cls.get_edge_coloring_for_regular(b3, start_color)
            b6 = cls.get_edge_coloring_for_regular(b4, start_color + deg // 2)
            cls._cpy_edge(b5, ret)
            cls._cpy_edge(b6, ret)
        return ret

    @classmethod
    def get_edge_coloring(cls, bipartite: Bipartite) -> Bipartite:
        """
        Coloring all edges on this bipartite.

        The input color attribute of edges in input bipartite would be ignored.

        Args:
        bipartite(Bipartite) : exec edge coloring on this bipartite

        Returns:
            Bipartite : return a new bipartite with color. Color is represented as an integer, starting from 0.
        """
        new_to_olds: Dict[int, List[int]]
        old_to_new: Dict[int, int]
        bipartite_regular: Bipartite
        new_to_olds, old_to_new, bipartite_regular = cls.get_regular(bipartite)
        colored_regular = cls.get_edge_coloring_for_regular(bipartite_regular, 0)
        # del bipartite_regular
        colored_bipartite = Bipartite(bipartite.left, bipartite.right)
        e = len(colored_regular.edges)
        new_eid_visited = [False for _ in range(e + 1)]
        """
        There are no parallel edges in the input bipartite graph. But parallel edges might be found
        in the regularized bipartite. We iterate over all old edges in the old bipartite to
        find their corresponding edges in the colored bipartite.
        """
        for node in bipartite.left:
            old_eid = bipartite.head[node]
            while old_eid != -1:
                old_edge = bipartite.edges[old_eid]

                old_s = node
                old_e = old_edge.end
                new_s = old_to_new[old_s]
                new_e = old_to_new[old_e]
                new_eid = colored_regular.head[new_s]
                while new_eid != -1:
                    new_edge = colored_regular.edges[new_eid]

                    if new_eid_visited[new_eid] or (not new_edge.valid) or (new_edge.end != new_e):
                        new_eid = new_edge.next
                        continue

                    new_eid_visited[new_eid] = True
                    new_eid_visited[new_eid ^ 1] = True

                    clr = new_edge.color
                    colored_bipartite.add_edge(
                        start=old_s,
                        end=old_e,
                        valid=True,
                        color=clr
                    )
                    colored_bipartite.add_edge(
                        start=old_e,
                        end=old_s,
                        valid=True,
                        color=clr
                    )

                    break

                old_eid = old_edge.next

        return colored_bipartite



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2019/12/13 2:20
# @Author  : Han Yu
# @File    : __version__.py

__title__ = "QuICT"
__version__ = "1.0.0"
__description__ = "quantum algorithm library"
__author__ = "ICT_thoery"
__email__ = "hanyu16@mails.ucas.ac.cn"
__url__ = "http://theory.ict.ac.cn/"



from typing import *
import numpy as np

from QuICT.core import *
from QuICT.core.gate import *
from QuICT.core.gate.backend import UniformlyRotation
from .utility import *


class ControlledUnitaryDecomposition(object):
    def __init__(self, include_phase_gate: bool = True, recursive_basis: int = 2):
        """
        Args:
            include_phase_gate(bool): Whether to include a phase gate to keep synthesized gate matrix the same
                as input. If set False, the output gates might have a matrix which has a factor shift to input:
                np.allclose(<matrix_of_return_gates> * factor, <input_matrix>).
            recursive_basis(int): Terminate recursion at which level. It could be set as 1 or 2, which would stop
                recursion when matrix is 2 or 4, respectively. When set as 2, the final step is done by KAK
                decomposition.
                Correctness of this algorithm is never influenced by recursive_basis.
        """
        self.include_phase_gate = include_phase_gate
        self.recursive_basis = recursive_basis

    def execute(self, u1: np.ndarray, u2: np.ndarray):
        """
        Transform a controlled-unitary matrix into CX gates and single qubit gates.
        A controlled-unitary is a block-diagonal unitary. Parameter u1 and u2 are
        the block diagonals.

        Args:
            u1(np.ndarray): Upper-left block diagonal.
            u2(np.ndarray): bottom-right block diagonal.
            include_phase_gate(bool): Whether to include a phase gate to keep synthesized gate matrix the same
                as input. If set False, the output gates might have a matrix which has a factor shift to input:
                np.allclose(<matrix_of_return_gates> * factor, <input_matrix>).
            mapping(List[int]): The order of input qubits. Mapping is a list of their labels from top to bottom.
            recursive_basis(int): Terminate recursion at which level. It could be set as 1 or 2, which would stop
                recursion when matrix is 2 or 4, respectively. When set as 2, the final step is done by KAK
                decomposition.
                Correctness of this algorithm is never influenced by recursive_basis.

        Returns:
            Union[Tuple[CompositeGate,None], Tuple[CompositeGate,complex]]: If inlclude_phase_gate==False,
                this function returns synthesized gates and a shift factor. Otherwise a tuple like (<gates>, None)
                is returned.
        """
        gates, shift = self.inner_cutrans_build_gate(u1, u2, self.recursive_basis)
        if self.include_phase_gate:
            gates = add_factor_shift_into_phase(gates, shift)
            return gates, None
        else:
            return gates, shift

    def _i_tensor_unitary(
        self,
        u: np.ndarray,
        recursive_basis: int,
        keep_left_diagonal: bool = False,
    ) -> Tuple[CompositeGate, complex]:
        """
        Transform (I_{2x2} tensor U) into gates. The 1st bit
        is under the identity transform.

        Args:
            u (np.ndarray): A unitary matrix.

        Returns:
            Tuple[CompositeGate, complex]: Synthesized gates and a phase factor.
        """

        gates: CompositeGate
        shift: complex
        # Dynamically import to avoid circulation.
        from .unitary_decomposition import UnitaryDecomposition
        UD = UnitaryDecomposition()
        gates, shift = UD.inner_utrans_build_gate(
            mat=u,
            recursive_basis=recursive_basis,
            keep_left_diagonal=keep_left_diagonal,
        )
        for gate in gates:
            for idx, _ in enumerate(gate.cargs):
                gate.cargs[idx] += 1
            for idx, _ in enumerate(gate.targs):
                gate.targs[idx] += 1

        return gates, shift

    def inner_cutrans_build_gate(
        self,
        u1: np.ndarray,
        u2: np.ndarray,
        recursive_basis: int = 2,
        keep_left_diagonal: bool = False,
    ) -> Tuple[CompositeGate, complex]:
        """
        Build gates from parameterized model without mapping

        Returns:
            Tuple[CompositeGate, complex]: Synthesized gates and factor shift.
        """
        qubit_num = 1 + int(round(np.log2(u1.shape[0])))
        v, d, w = quantum_shannon_decompose(u1, u2)
        shift: complex = 1.0

        # diag(u1, u2) == diag(v, v) @ diag(d, d_dagger) @ diag(w, w)
        # diag(v, v)
        v_gates, _shift = self._i_tensor_unitary(v, recursive_basis, keep_left_diagonal=True)
        shift *= _shift

        # diag(d, d_dagger)
        angle_list = []
        for i in range(d.shape[0]):
            s = d[i, i]
            theta = -2 * np.log(s) / 1j
            angle_list.append(theta)

        URz = UniformlyRotation(GateType.rz)
        reversed_rz = URz.execute(angle_list)
        reversed_rz & [(i + 1) % qubit_num for i in range(qubit_num)]

        # diag(w, w)
        if recursive_basis == 2:
            forwarded_d_gate: BasicGate = v_gates.gates.pop(0)
            forwarded_mat = forwarded_d_gate.matrix
            for i in range(0, w.shape[0], 4):
                for k in range(4):
                    w[i + k, :] *= forwarded_mat[k, k]

        w_gates, _shift = self._i_tensor_unitary(w, recursive_basis, keep_left_diagonal=keep_left_diagonal)
        shift *= _shift

        gates = CompositeGate()
        gates.extend(w_gates)
        gates.extend(reversed_rz)
        gates.extend(v_gates)

        return gates, shift



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/1/17 8:44
# @Author  : Han Yu, Li Kaiqi
# @File    : qubit.py
import random
from typing import Union

from QuICT.core.utils import unique_id_generator
from QuICT.tools.exception.core import TypeError, ValueError, IndexExceedError, QubitMeasureError


class Qubit(object):
    """ Implement a quantum bit

    Qubit is the basic unit of quantum circuits, it will appear with some certain circuit.

    Attributes:
        id(int): the unique identity code of a qubit, which is generated globally.
        measured(int):
            the measure result of the qubit.
            After apply measure gate on the qubit, the measured will be 0 or 1,
            otherwise raise an exception
        prob(float):
            the probability of measure result to be 1, which range in [0, 1].
            After apply measure gate on the qubit, this attribute can be read,
            otherwise raise an exception
        historical_measured(list):
            Record all measured result of current qubits.
    """

    @property
    def id(self):
        return self._id

    @property
    def measured(self) -> int:
        return self._measured

    @measured.setter
    def measured(self, measured):
        self._measured = measured
        if self._measured is not None:
            self._historical_measured.append(self._measured)

    @property
    def historical_measured(self):
        return self._historical_measured

    @property
    def prob(self) -> float:
        return self._prob

    @prob.setter
    def prob(self, prob):
        self._prob = prob

    def __init__(self, prob: float = random.random()):
        """ initial a qubit with a circuit

        Args:
            circuit(Circuit): the circuit the qubit attaches to
        """
        self._id = unique_id_generator()
        self._measured = None
        self._prob = prob
        self._historical_measured = []

    def __str__(self):
        """ string describe of the qubit

        Returns:
            str: a simple describe
        """
        return f"qubit id: {self.id}"

    def __int__(self):
        """ int value of the qubit(measure result)

        Returns:
            int: measure result

        Raises:
            The qubit has not be measured.
        """
        if self.measured is None:
            raise QubitMeasureError(f"The qubit {self.id} has not be measured")

        return self.measured

    def __bool__(self):
        """ int value of the qubit(measure result)

        Returns:
            bool: measure result

        Raises:
            The qubit has not be measured.
        """
        return bool(int(self))

    def reset(self):
        """ Reset self qubit status. """
        self._historical_measured = []
        self._measured = None


class Qureg(list):
    """ Implement a quantum register

    Qureg is a list of Qubits, which is a subClass of list.

    Attributes:
        qubits([Qubits]): the list of qubits.
    """
    def __init__(self, qubits=None):
        """ initial a qureg with qubit(s)

        Args:
            qubits: the qubits which make up the qureg, it can have below form,
                1) int
                2) qubit
                3) [qubits/quregs]
        """
        super().__init__()
        if qubits is None:
            return

        if isinstance(qubits, int):
            for _ in range(qubits):
                self.append(Qubit())
        elif isinstance(qubits, Qubit):
            self.append(qubits)
        elif isinstance(qubits, list):
            for qubit in qubits:
                if isinstance(qubit, Qubit):
                    self.append(qubit)
                elif isinstance(qubit, Qureg):
                    for qbit in qubit:
                        self.append(qbit)
                else:
                    raise TypeError("Qureg.qubits", "int/Qubit/list<Qubit/Qureg>", type(qubit))
        else:
            raise TypeError("Qureg.qubits", "int/Qubit/list<Qubit/Qureg>", type(qubit))

    def __call__(self, indexes: object):
        """ get a smaller qureg from this qureg

        Args:
            indexes: the indexes passed in, it can have follow form:
                1) int
                2) list<int>
        Returns:
            Qubit[s]: the qureg correspond to the indexes
        """
        if isinstance(indexes, int):
            return Qureg(self[indexes])

        return self[indexes]

    def __getitem__(self, item):
        """ to fit the slice operator, overloaded this function.

        get a smaller qureg/qubit from this qureg

        Args:
            item(int/slice): slice passed in.

        Return:
            Qubit/Qureg: the result or slice
        """
        if isinstance(item, int):
            return super().__getitem__(item)

        qureg = Qureg()
        if isinstance(item, slice):
            qureg_list = super().__getitem__(item)
            for qubit in qureg_list:
                qureg.append(qubit)
        elif isinstance(item, list) or isinstance(item, tuple):
            for idx in item:
                assert isinstance(idx, int), TypeError("Qureg.getitem.item.value", "int", type(idx))
                if idx < 0 or idx > len(self):
                    raise IndexExceedError("Qureg.getitem", [0, len(self)], idx)

                qureg.append(self[idx])
        else:
            raise TypeError("Qureg.getitem", "int/list[int]/slice", type(item))

        return qureg

    def __int__(self):
        """ the value of the register

        Return the value of the register if all qubits have been measured.
        Note that the compute mode is BigEndian.

        Returns:
            int: the value of the register

        Raises:
            Exception: some qubit has not be measured
        """
        value = 0
        for qubit in self:
            if qubit.measured is None:
                raise QubitMeasureError(f"The qubit {qubit.id} has not be measured")

            value <<= 1
            if qubit.measured == 1:
                value += 1

        return value

    def __str__(self):
        """ the string of the value of the register

        Returns:
            str: the value of the qureg
        """
        bit_idx = "{0:0b}".format(self.__int__())
        bit_idx.zfill(len(self))

        return bit_idx

    def __add__(self, other):
        """ get a combined qureg with this qureg and other qureg

        Args:
            other(Qureg): qureg to be added.

        Return:
            Qureg: the result or slice
        """
        return Qureg([self, other])

    def __iadd__(self, other):
        """ get a combined qureg with this qureg and other qureg

        Args:
            other(Qureg): qureg to be added.

        Return:
            Qureg: the result or slice
        """
        if isinstance(other, Qubit):
            self.append(other)
        elif isinstance(other, Qureg):
            for q in other:
                self.append(q)
        else:
            raise TypeError("Qureg.iadd", "Qureg/Qubit", type(other))

        return self

    def __eq__(self, other):
        """
        check two qureg is same or not. Iff all qubits in two qureg are same will
        return True; otherwise, return False.

        Args:
            other(Qureg): qureg to be checked.
        """
        assert isinstance(other, Qureg), TypeError("Qureg.eq", "Qureg", type(other))
        if not len(other) == len(self):
            return False

        current_qubit_ids = [qubit.id for qubit in self]
        for qubit in other:
            if qubit.id not in current_qubit_ids:
                return False

        return True

    def diff(self, other):
        """ return different qubits between two quregs

        Args:
            other (Qureg): The compare qureg

        Returns:
            Qureg: The qureg with different qubits
        """
        qubit_ids = [qubit.id for qubit in self]
        diff_qubit = []

        for qubit in other:
            if qubit.id not in qubit_ids:
                diff_qubit.append(qubit)

        return Qureg(diff_qubit)

    def index(self, qubit: Union[str, Qubit]):
        if isinstance(qubit, Qubit):
            return super().index(qubit)

        for idx, item in enumerate(self):
            if item.id == qubit:
                return idx

        raise ValueError("Qureg.index.qubit", "within current Qureg", "qubit is not")

    def reset_qubits(self):
        """ Reset all qubits' status. """
        for qubit in self:
            qubit.reset()



"""
Synthesize a Clifford circuit unidirectionally or bidirectionally
"""

import copy
import random
import multiprocessing as mp

import numpy as np

from QuICT.core import Circuit
from QuICT.core.gate import CompositeGate, GateType
from QuICT.qcda.utility import PauliOperator, OutputAligner


class CliffordUnidirectionalSynthesizer(object):
    """
    Construct L_1,…,L_n such that C = L_1…L_j C_j, where C_j acts trivially on the first j qubits.
    By induction the original Clifford circuit C is synthesized.

    Reference:
        https://arxiv.org/abs/2105.02291
    """
    def __init__(self, strategy='greedy'):
        """
        Args:
            strategy(str, optional): strategy of choosing qubit for each step, in ['greedy', 'random']
        """
        assert strategy in ['greedy', 'random'],\
            ValueError('strategy of choosing qubit could only be "greedy" or "random"')
        self.strategy = strategy

    @OutputAligner()
    def execute(self, gates: CompositeGate):
        """
        Args:
            gates(Circuit/CompositeGate): the Clifford Circuit/CompositeGate to be synthesized

        Returns:
            CompositeGate: the synthesized Clifford CompositeGate
        """
        width = gates.width()
        if isinstance(gates, Circuit):
            gates = CompositeGate(gates=gates.gates)
        assert isinstance(gates, CompositeGate),\
            TypeError('Invalid input(Circuit/CompositeGate)')
        for gate in gates.gates:
            assert gate.is_clifford(), TypeError('Only Clifford gates here')

        def gates_next(gates: CompositeGate, disentangler: CompositeGate):
            gates_next = disentangler.inverse()
            gates_next.extend(gates)
            return gates_next

        gates_syn = CompositeGate()
        not_disentangled = list(range(width))
        if self.strategy == 'greedy':
            while not_disentangled:
                cnot_min = np.inf
                disentangler_min = None
                qubit_min = None
                for qubit in not_disentangled:
                    disentangler = self.disentangle_one_qubit(gates, width, qubit)
                    if disentangler.count_2qubit_gate() < cnot_min:
                        cnot_min = disentangler.count_2qubit_gate()
                        disentangler_min = disentangler
                        qubit_min = qubit
                gates_syn.extend(disentangler_min)
                gates = gates_next(gates, disentangler_min)
                not_disentangled.remove(qubit_min)
        else:
            while not_disentangled:
                qubit = random.choice(not_disentangled)
                disentangler = self.disentangle_one_qubit(gates, width, qubit)
                gates_syn.extend(disentangler)
                gates = gates_next(gates, disentangler)
                not_disentangled.remove(qubit)

        return gates_syn

    @staticmethod
    def disentangle_one_qubit(gates: CompositeGate, width: int, target: int):
        """
        Disentangle the target qubit from gates, i.e. for CompositeGate C, give the CompositeGate L
        such that L^-1 C acts trivially on the target qubit.

        Args:
            gates(CompositeGate): the CompositeGate to be disentangled
            width(int): the width of the operators
            target(int): the target qubit to be disentangled from gates

        Returns:
            CompositeGate: the disentangler
        """
        # Create X_j, Z_j
        pauli_x = PauliOperator([GateType.id for _ in range(width)])
        pauli_z = PauliOperator([GateType.id for _ in range(width)])
        pauli_x.operator[target] = GateType.x
        pauli_z.operator[target] = GateType.z

        # Compute C X_j C^-1 and C Z_j C^-1
        for gate in gates.inverse():
            pauli_x.conjugate_act(gate)
            pauli_z.conjugate_act(gate)

        return PauliOperator.disentangler(pauli_x, pauli_z, target)


class CliffordBidirectionalSynthesizer(object):
    """
    Construct L_1,…,L_n,R_1,…,R_n such that C = L_1…L_j C_j R_j…R_1,  where C_j acts trivially
    on the first j qubits.
    By induction the original Clifford circuit C is synthesized.

    Reference:
        https://arxiv.org/abs/2105.02291
    """
    def __init__(self, qubit_strategy='greedy', pauli_strategy='random',
                 shots=1, multiprocess=False, process=4, chunksize=64):
        """
        Args:
            qubit_strategy(str, optional): strategy of choosing qubit for each step, in ['greedy', 'random']
            pauli_strategy(str, optional): strategy of choosing PauliOperator for each step, in ['greedy', 'random']
            shots(int, optional): if pauli_strategy is random, shots of random
            multiprocess(bool, optional): whether to use the multiprocessing accelaration
            process(int, optional): the number of processes in a pool
            chunksize(int, optional): iteration dealt with in a process
        """
        assert qubit_strategy in ['greedy', 'random'],\
            ValueError('strategy of choosing qubit could only be "greedy" or "random"')
        assert pauli_strategy in ['brute_force', 'random'],\
            ValueError('strategy of choosing PauliOperator could only be "brute_force" or "random"')
        self.qubit_strategy = qubit_strategy
        self.pauli_strategy = pauli_strategy
        self.shots = shots
        self.multiprocess = multiprocess
        self.process = process
        self.chunksize = chunksize

    @OutputAligner()
    def execute(self, gates: CompositeGate):
        """
        Args:
            gates(Circuit/CompositeGate): the Clifford Circuit/CompositeGate to be synthesized

        Returns:
            CompositeGate: the synthesized Clifford CompositeGate
        """
        width = gates.width()
        if isinstance(gates, Circuit):
            gates = CompositeGate(gates=gates.gates)
        assert isinstance(gates, CompositeGate),\
            TypeError('Invalid input(Circuit/CompositeGate)')
        for gate in gates.gates:
            assert gate.is_clifford(), TypeError('Only Clifford gates here')

        def gates_next(gates: CompositeGate, left: CompositeGate, right: CompositeGate):
            gates_next = left.inverse()
            gates_next.extend(gates)
            gates_next.extend(right)
            return gates_next

        gates_left = CompositeGate()
        gates_right = CompositeGate()
        not_disentangled = list(range(width))

        if self.qubit_strategy == 'greedy':
            while not_disentangled:
                cnot_min = np.inf
                left_min = None
                right_min = None
                qubit_min = None
                for qubit in not_disentangled:
                    cnot_cnt, left, right = self._minimum_over_pauli(gates, width, qubit, not_disentangled)
                    if cnot_cnt < cnot_min:
                        cnot_min = cnot_cnt
                        left_min = left
                        right_min = right
                        qubit_min = qubit
                gates_left.extend(left_min)
                gates_right.left_extend(right_min.inverse())
                gates = gates_next(gates, left_min, right_min)
                not_disentangled.remove(qubit_min)
        else:
            while not_disentangled:
                qubit = random.choice(not_disentangled)
                _, left, right = self._minimum_over_pauli(gates, width, qubit, not_disentangled)
                gates_left.extend(left)
                gates_right.left_extend(right.inverse())
                gates = gates_next(gates, left, right)
                not_disentangled.remove(qubit)

        gates_left.extend(gates_right)
        return gates_left

    @staticmethod
    def _insert_identity(p: PauliOperator, width: int, not_disentangled: list):
        # Operators on the disentangled qubits must be I
        for i in range(width):
            if i not in not_disentangled:
                p.operator.insert(i, GateType.id)

    @classmethod
    def _brute_force_iterator(cls, gates: CompositeGate, width: int, qubit: int, not_disentangled: list):
        for p1 in PauliOperator.iterator(len(not_disentangled)):
            cls._insert_identity(p1, width, not_disentangled)
            for p2 in PauliOperator.iterator(len(not_disentangled)):
                cls._insert_identity(p2, width, not_disentangled)
                # Only anti-commutative pairs
                if p1.commute(p2):
                    continue
                yield gates, qubit, p1, p2

    @classmethod
    def _random_iterator(cls, gates: CompositeGate, width: int, qubit: int, not_disentangled: list, shots: int):
        for _ in range(shots):
            p1, p2 = PauliOperator.random_anti_commutative_pair(len(not_disentangled))
            cls._insert_identity(p1, width, not_disentangled)
            cls._insert_identity(p2, width, not_disentangled)
            yield gates, qubit, p1, p2

    def _minimum_over_pauli(self, gates: CompositeGate, width: int, qubit: int, not_disentangled: list):
        cnot_min = np.inf
        left_min = None
        right_min = None
        if self.pauli_strategy == 'brute_force':
            iterator = self._brute_force_iterator(gates, width, qubit, not_disentangled)
        if self.pauli_strategy == 'random':
            iterator = self._random_iterator(gates, width, qubit, not_disentangled, self.shots)
        if self.multiprocess:
            pool = mp.Pool(self.process)
            result = pool.starmap_async(self.disentangle_one_qubit,
                                        iterable=iterator,
                                        chunksize=self.chunksize)
            for left, right in result.get():
                if left.count_2qubit_gate() + right.count_2qubit_gate() < cnot_min:
                    cnot_min = left.count_2qubit_gate() + right.count_2qubit_gate()
                    left_min = left
                    right_min = right
            pool.close()
            pool.join()
        else:
            for gates, qubit, p1, p2 in iterator:
                left, right = self.disentangle_one_qubit(gates, qubit, p1, p2)
                if left.count_2qubit_gate() + right.count_2qubit_gate() < cnot_min:
                    cnot_min = left.count_2qubit_gate() + right.count_2qubit_gate()
                    left_min = left
                    right_min = right
        return cnot_min, left_min, right_min

    @staticmethod
    def disentangle_one_qubit(gates: CompositeGate, target: int, p1: PauliOperator, p2: PauliOperator):
        """
        Disentangle the target qubit from gates, i.e. for CompositeGate C, give the CompositeGate L, R
        such that L^-1 C R^-1 acts trivially on the target qubit.

        Args:
            gates(CompositeGate): the CompositeGate to be disentangled
            target(int): the target qubit to be disentangled from gates
            p1(PauliOperator): PauliOperator P
            p2(PauliOperator): PauliOperator P'

        Returns:
            CompositeGate, CompositeGate: the left and right disentangler
        """
        # Using the notation in the paper
        o1 = copy.deepcopy(p1)
        o2 = copy.deepcopy(p2)

        # Compute C P C^-1 and C P' C^-1
        for gate in gates.inverse():
            o1.conjugate_act(gate)
            o2.conjugate_act(gate)

        return PauliOperator.disentangler(o1, o2, target), PauliOperator.disentangler(p1, p2, target)



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/12 11:37
# @Author  : Han Yu
# @File    : MCT_one_aux.py

from QuICT.core import *
from QuICT.core.gate import *
from .mct_linear_dirty_aux import MCTLinearHalfDirtyAux


class MCTOneAux(object):
    def execute(self, n):
        """ Decomposition of n-qubit Toffoli gates with one ancillary qubit and linear circuit complexity

        He Y, Luo M X, Zhang E, et al.
        Decompositions of n-qubit Toffoli gates with linear circuit complexity[J].
        International Journal of Theoretical Physics, 2017, 56(7): 2350-2361.

        Args:
            n(int): the number of used qubit, which is (n + 2) for n-qubit Toffoli gates

        Returns:
            CompositeGate: the result of Decomposition
        """
        n = n - 1
        gates = CompositeGate()
        qubit_list = list(range(n + 1))
        with gates:
            if n == 3:
                CCX & qubit_list[:3]
                return gates
            elif n == 2:
                CX & qubit_list[:2]
                return gates
            k = n // 2 + n % 2

            MCT_half_dirty = MCTLinearHalfDirtyAux()
            half_dirty_gates = MCT_half_dirty.execute(k, n + 1)

            qubit_remain = qubit_list[k:k + n // 2] + qubit_list[:k] + [qubit_list[-1]]
            if n // 2 < 1:
                raise Exception("there must be at least one control bit")
            controls_remain = [qubit_remain[i] for i in range(n // 2)]
            auxs_remain = [qubit_remain[i] for i in range(n // 2, n)]
            gates_remain = MCT_half_dirty.assign_qubits(n + 1, n // 2, controls_remain, auxs_remain, n)

            half_dirty_gates | gates
            H & qubit_list[-2]
            S & qubit_list[-1]
            gates_remain | gates
            S_dagger & qubit_list[-1]
            half_dirty_gates | gates
            S & qubit_list[-1]
            gates_remain | gates
            H & qubit_list[-2]
            S_dagger & qubit_list[-1]

        return gates



from QuICT.tools.exception import QuICTException


class SampleBeforeRunError(QuICTException):
    """ Simulation sample before run error. """
    def __init__(self, msg: str):
        super().__init__(2001, msg)


class GateTypeNotImplementError(QuICTException):
    """ The given gate is not implemented in simulation. """
    def __init__(self, msg: str = None):
        super().__init__(2002, msg)


class SimulationMatrixError(QuICTException):
    """ Simulation Matrix Error"""
    def __init__(self, msg: str):
        super().__init__(2003, msg)


class StateVectorUnmatchedError(QuICTException):
    """ The State Vector is not matched the simulation. """
    def __init__(self, msg: str):
        super().__init__(2004, msg)


class GateAlgorithmNotImplementError(QuICTException):
    """ The algorithm of simulate quantum gate is not implemented. """
    def __init__(self, msg: str = None):
        super().__init__(2005, msg)


class UnitaryMatrixUnmatchedError(QuICTException):
    """ The State Vector is not matched the simulation. """
    def __init__(self, msg: str):
        super().__init__(2006, msg)


class SimulatorOptionsUnmatchedError(QuICTException):
    """ The backend and optional arguments is not matched in the simulation. """
    def __init__(self, msg: str):
        super().__init__(2007, msg)



from ._operator import Operator


class Multiply(Operator):
    def __init__(self, value):
        super().__init__(targets=1)
        self._value = value

    @property
    def value(self):
        return self._value

    def qasm(self):
        return f"{self._value}"


class SpecialGate(Operator):
    @property
    def proxy_idx(self):
        return self._proxy_idx

    @property
    def type(self):
        return self._type

    def __init__(self, type, targ: list, proxy_idx: int = -1):
        super().__init__(targets=1)
        self._type = type
        self._proxy_idx = proxy_idx
        self.targs = targ

    def qasm(self):
        return f"{self._type}, {self.targs}, {self._proxy_idx}."

    def __str__(self):
        return f"{self._type}, {self.targs}, {self._proxy_idx}."



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/11/5 10:55
# @Author  : Han Yu
# @File    : VBE.py

from numpy import gcd

from QuICT.core import Circuit
from QuICT.core.gate import CompositeGate, X, CX, CCX, Swap


def Inverse(a, N):
    """ Inversion of a in (mod N)

    Args:
        a(int): the parameter a
        N(int): the parameter N

    """
    for i in range(N):
        if i * a % N == 1:
            return i
    return None


def Set(qreg, N):
    """ Set the qreg as N, using X gates on specific qubits

    Args:
        qreg(Qureg): the qureg to be set
        N(int): the parameter N

    """

    string = bin(N)[2:]
    n = len(qreg)
    m = len(string)
    if m > n:
        print('When set qureg as N=%d, N exceeds the length of qureg n=%d, thus is truncated' % (N, n))

    for i in range(min(n, m)):
        if string[m - 1 - i] == '1':
            X | qreg[n - 1 - i]


def ControlSet(control, qreg, N):
    """ Set the qreg as N, using CX gates on specific qubits

    Args:
        control(Qubit): qubit of control bits
        qreg(Qureg): the qureg to be set
        N(int): the parameter N

    """
    string = bin(N)[2:]
    n = len(qreg)
    m = len(string)
    if m > n:
        print(
            f'When cset qureg as N={N}, N exceeds the length of qureg n={n}, thus is truncated')

    for i in range(min(n, m)):
        if string[m - 1 - i] == '1':
            CX | (control, qreg[n - 1 - i])


def CControlSet(control1, control2, qreg, N):
    """ Set the qreg as N, using CCX gates on specific qubits

    Args:
        control1(Qubit): 1st qubit of control bits
        control2(Qubit): 2nd qubit of control bits
        qreg(Qureg): the qureg to be set
        N(int): the parameter N

    """
    string = bin(N)[2:]
    n = len(qreg)
    m = len(string)
    if m > n:
        print(
            f'When ccset qureg as N={N}, N exceeds the length of qureg n={n}, thus is truncated')

    for i in range(min(n, m)):
        if string[m - 1 - i] == '1':
            CCX | (control1, control2, qreg[n - 1 - i])


def Carry(c_in, a, b, c_out):
    """ Carry for one bit plus

    (c_in,a,b,c_out=0) -> (c_in,a,b,c_out')

    Args:
        c_in(Qubit): c_in
        a(Qubit): a
        b(Qubit): b
        c_out(Qubit): c_out

    """
    CCX | (a, b, c_out)
    CX | (a, b)
    CCX | (c_in, b, c_out)


def ReverseCarry(c_in, a, b, c_out):
    """ the inverse of Carry

    Args:
        c_in(Qubit): c_in
        a(Qubit): a
        b(Qubit): b
        c_out(Qubit): c_out
    """
    CCX | (c_in, b, c_out)
    CX | (a, b)
    CCX | (a, b, c_out)


def Sum(c_in, a, b):
    """ Sum circuit

    (c_in,a,b) -> (c_in,a,b'=a+b+c_in)

    Args:
        c_in(Qubit): c_in
        a(Qubit): a
        b(Qubit): b
    """
    CX | (a, b)
    CX | (c_in, b)


def ReverseSum(c_in, a, b):
    """ Reverse of Sum

    Args:
        c_in(Qubit): c_in
        a(Qubit): a
        b(Qubit): b
    """
    CX | (c_in, b)
    CX | (a, b)


def PlainAdder(a, b, c, overflow):
    """ store a + b in b

    (a,b,c=0,overflow=0) -> (a,b'=a+b,c,overflow')


    Args:
        a(Qureg): the qureg stores a, length is n
        b(Qureg): the qureg stores b, length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1

    """
    n = len(a)

    for i in range(n - 1):
        Carry(c[n - 1 - i], a[n - 1 - i], b[n - 1 - i], c[n - 2 - i])
    Carry(c[0], a[0], b[0], overflow)

    CX | (a[0], b[0])
    Sum(c[0], a[0], b[0])

    for i in range(n - 1):
        ReverseCarry(c[1 + i], a[1 + i], b[1 + i], c[i])
        Sum(c[1 + i], a[1 + i], b[1 + i])


def ReversePlainAdder(a, b, c, overflow):
    """ the inverse of plainAdder

    Args:
        a(Qureg): the qureg stores a, length is n
        b(Qureg): the qureg stores b, length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1

    """
    n = len(a)

    for i in range(n - 1):
        ReverseSum(c[n - 1 - i], a[n - 1 - i], b[n - 1 - i])
        Carry(c[n - 1 - i], a[n - 1 - i], b[n - 1 - i], c[n - 2 - i])

    ReverseSum(c[0], a[0], b[0])
    CX | (a[0], b[0])

    ReverseCarry(c[0], a[0], b[0], overflow)
    for i in range(n - 1):
        ReverseCarry(c[1 + i], a[1 + i], b[1 + i], c[i])


def AdderMod(N, a, b, c, overflow, qubit_N, t):
    """ store (a+b) mod N in b

    (a,b,c=0,overflow=0,qubit_N=0,t=0) ->
     (a,b'=(a+b)mod N,c,overflow',qubit_N=0,t=0)

    Args:
        N(int): the parameter N
        a(Qureg): the qureg stores a, length is n
        b(Qureg): the qureg stores b, length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1
        qubit_N(Qureg): the ancillary qubits, length is n
        t(Qureg): the ancillary qubits, length is 1

    """
    n = len(qubit_N)

    Set(qubit_N, N)
    PlainAdder(a, b, c, overflow)
    for i in range(n):
        Swap | (a[i], qubit_N[i])
    ReversePlainAdder(a, b, c, overflow)
    X | overflow
    CX | (overflow, t)
    X | overflow
    ControlSet(t, a, N)
    PlainAdder(a, b, c, overflow)
    ControlSet(t, a, N)
    for i in range(n):
        Swap | (a[i], qubit_N[i])
    ReversePlainAdder(a, b, c, overflow)
    CX | (overflow, t)
    PlainAdder(a, b, c, overflow)
    Set(qubit_N, N)


def ReverseAdderMod(N, a, b, c, overflow, qubit_N, t):
    """ Reverse of AdderMod """
    n = len(qubit_N)

    Set(qubit_N, N)
    ReversePlainAdder(a, b, c, overflow)
    CX | (overflow, t)
    PlainAdder(a, b, c, overflow)
    for i in range(n):
        Swap | (a[i], qubit_N[i])
    ControlSet(t, a, N)
    ReversePlainAdder(a, b, c, overflow)
    ControlSet(t, a, N)
    X | overflow
    CX | (overflow, t)
    X | overflow
    PlainAdder(a, b, c, overflow)
    for i in range(n):
        Swap | (a[i], qubit_N[i])
    ReversePlainAdder(a, b, c, overflow)
    Set(qubit_N, N)


def MulAddMod(a, N, x, qubit_a, b, c, overflow, qubit_N, t):
    """ store b + x*a mod N in b

    (x,qubit_a=0,b,c=0,overflow=0,qubit_N=0,t=0) ->
    (x,qubit_a=0,b'=b+a*x mod N,c,overflow',qubit_N,t)

    Args:
        a(int): the parameter a
        N(int): the parameter N
        x(Qureg): the qureg stores x, length is m
        qubit_a(Qureg): the ancillary qubits, length is n
        b(Qureg): length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1
        qubit_N(Qureg): the ancillary qubits, length is n
        t(Qureg): the ancillary qubits, length is 1
    """

    n = len(qubit_N)

    for i in range(n):
        ControlSet(x[n - 1 - i], qubit_a, a)
        AdderMod(N, qubit_a, b, c, overflow, qubit_N, t)
        ControlSet(x[n - 1 - i], qubit_a, a)
        a = (a * 2) % N


def ControlMulMod(a, N, control, x, qubit_a, b, c, overflow, qubit_N, t):
    """ store x*(a^control) mod N in b

    (control,x,qubit_a=0,b=0,c=0,overflow=0,qubit_N=0,t=0) ->
     (control,x,qubit_a=0,b'=(a**control)*x mod N,c,overflow',qubit_N,t)

    Args:
        a(int): the parameter a
        N(int): the parameter N
        control(Qureg): the control qureg to store the result, length is 1
        x(Qureg): the qureg stores x, length is m
        qubit_a(Qureg): the ancillary qubits, length is n
        b(Qureg): length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1
        qubit_N(Qureg): the ancillary qubits, length is n
        t(Qureg): the ancillary qubits, length is 1
    """

    n = len(qubit_N)

    for i in range(n):
        CControlSet(control, x[n - 1 - i], qubit_a, a)
        AdderMod(N, qubit_a, b, c, overflow, qubit_N, t)
        CControlSet(control, x[n - 1 - i], qubit_a, a)
        a = (a * 2) % N

    X | control
    for i in range(n):
        CCX | (control, x[i], b[i])
    X | control


def ReverseControlMulMod(a, N, control, x, qubit_a, b, c, overflow, qubit_N, t):
    """ Reverse of ControlMulMod

    Args:
        a(int): the parameter a
        N(int): the parameter N
        control(Qureg): the control qureg to store the result, length is 1
        x(Qureg): the qureg stores x, length is m
        qubit_a(Qureg): the ancillary qubits, length is n
        b(Qureg): length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1
        qubit_N(Qureg): the ancillary qubits, length is n
        t(Qureg): the ancillary qubits, length is 1
    """
    n = len(qubit_N)

    a_list = []
    for i in range(n):
        a_list.append(a)
        a = a * 2 % N

    X | control
    for i in range(n):
        CCX | (control, x[i], b[i])
    X | control

    for i in range(n):
        CControlSet(control, x[i], qubit_a, a_list[n - 1 - i])
        ReverseAdderMod(N, qubit_a, b, c, overflow, qubit_N, t)
        CControlSet(control, x[i], qubit_a, a_list[n - 1 - i])


def ExpMod(a, N, x, result, qubit_a, b, c, overflow, qubit_N, t):
    """ store a^x mod N in result

    Args:
        a(int): the parameter a
        N(int): the parameter N
        x(Qureg): the qureg stores x, length is m
        result(Qureg): the qureg to store the result
        qubit_a(Qureg): the ancillary qubits, length is n
        b(Qureg): the ancillary qubits, length is n
        c(Qureg): the ancillary qubits, length is n
        overflow(Qureg): the ancillary qubits, length is 1
        qubit_N(Qureg): the ancillary qubits, length is n
        t(Qureg): the ancillary qubits, length is 1
    """
    m = len(x)
    n = len(qubit_N)
    a_inv = Inverse(a, N)

    for i in range(m):
        ControlMulMod(a, N, x[m - 1 - i], result,
                      qubit_a, b, c, overflow, qubit_N, t)
        for j in range(n):
            Swap | (result[j], b[j])
        ReverseControlMulMod(
            a_inv, N, x[m - 1 - i], result, qubit_a, b, c, overflow, qubit_N, t)
        a = (a ** 2) % N
        a_inv = (a_inv ** 2) % N


class VBEAdder(object):
    @staticmethod
    def execute(n):
        """ a circuit calculate a+b, a and b are gotten from some qubits.

        (a,b,c=0,overflow) -> (a,b'=a+b,c=0,overflow')

        Quregs:
            a: the qureg stores a, length is n,
            b: the qureg stores b, length is n,
            c: the clean ancillary qubits, length is n,
            overflow: the dirty ancillary qubits, length is 1,
                            flips when overflows.

        Quantum Networks for Elementary Arithmetic Operations
        http://arxiv.org/abs/quant-ph/9511018v1
        """

        circuit = Circuit(3 * n + 1)
        qubit_a = circuit([i for i in range(n)])
        qubit_b = circuit([i for i in range(n, 2 * n)])
        qubit_c = circuit([i for i in range(2 * n, 3 * n)])
        qubit_overflow = circuit(3 * n)

        PlainAdder(qubit_a, qubit_b, qubit_c, qubit_overflow)

        return CompositeGate(gates=circuit.gates)


class VBEAdderMod(object):
    @staticmethod
    def execute(N, n):
        """ a circuit calculate (a+b) mod N.
        N are inherently designed in the circuit.

        (a,b,c=0,overflow=0,t=0,N) -> (a,b'=(a+b)%N,c=0,overflow=0,t=0,N)

        Quregs:
            a: the qureg stores a, length is n,
            b: the qureg stores b, length is n,
            c: the clean ancillary qubits, length is n,
            overflow: the clean ancillary qubits, length is 1,
            t: the clean ancillary qubits, length is 1.
            N: the qureg stores N, length is n,

        Quantum Networks for Elementary Arithmetic Operations
        http://arxiv.org/abs/quant-ph/9511018v1
        """

        circuit = Circuit(4 * n + 2)
        qubit_a = circuit([i for i in range(n)])
        qubit_b = circuit([i for i in range(n, 2 * n)])
        qubit_c = circuit([i for i in range(2 * n, 3 * n)])
        qubit_N = circuit([i for i in range(3 * n, 4 * n)])
        qubit_overflow = circuit(4 * n)
        qubit_t = circuit(4 * n + 1)

        AdderMod(N, qubit_a, qubit_b, qubit_c,
                 qubit_overflow, qubit_N, qubit_t)

        return CompositeGate(gates=circuit.gates)


class VBEMulAddMod(object):
    @staticmethod
    def execute(a, N, n, m):
        """ a circuit calculate b + x*a mod N.
        x are gotten from some qubits, a and N are inherently designed in the circuit.

        (x,b,qubit_a=0,c=0,overflow=0,qubit_N=0,t=0) ->
        (x,b'=b+a*x mod N,qubit_a,c,overflow,qubit_N,t)

        Quregs:
            x: the qureg stores x, length is m,
            b: the qureg stores b, length is n,
            qubit_a: the clean ancillary qubits, length is n,
            c: the clean ancillary qubits, length is n,
            overflow: the clean ancillary qubit, length is 1,
            qubit_N: the clean ancillary qubits, length is n,
            t: the clean ancillary qubit, length is 1.

        Quantum Networks for Elementary Arithmetic Operations
        http://arxiv.org/abs/quant-ph/9511018v1
        """

        circuit = Circuit(4 * n + m + 2)
        qubit_x = circuit([i for i in range(m)])
        qubit_a = circuit([i for i in range(m, n + m)])
        qubit_b = circuit([i for i in range(n + m, 2 * n + m)])
        qubit_c = circuit([i for i in range(2 * n + m, 3 * n + m)])
        qubit_overflow = circuit(3 * n + m)
        qubit_N = circuit([i for i in range(3 * n + m + 1, 4 * n + m + 1)])
        qubit_t = circuit(4 * n + m + 1)

        MulAddMod(a, N, qubit_x, qubit_a, qubit_b, qubit_c,
                  qubit_overflow, qubit_N, qubit_t)

        return CompositeGate(gates=circuit.gates)


class VBEExpMod(object):
    @staticmethod
    def execute(a, N, n, m):
        """ give parameters to the VBE
        Args:
            n(int): number of qubits of N
            m(int): number of qubits of x
            a(int): a
            N(int): N
        Returns:
            CompositeGate: the model filled by parameters.
        """
        if N <= 2:
            raise Exception("modulus should be great than 2")
        if gcd(a, N) != 1:
            raise Exception("a and N should be co-prime")
        # n = int(floor(log2(N))) + 1

        circuit = Circuit(m + 5 * n + 2)
        qubit_x = circuit([i for i in range(m)])
        qubit_r = circuit([i for i in range(m, m + n)])
        qubit_a = circuit([i for i in range(m + n, m + 2 * n)])
        qubit_b = circuit([i for i in range(m + 2 * n, m + 3 * n)])
        qubit_c = circuit([i for i in range(m + 3 * n, m + 4 * n)])

        overflow = circuit(m + 4 * n)
        qubit_N = circuit([i for i in range(m + 4 * n + 1, m + 5 * n + 1)])
        t = circuit(m + 5 * n + 1)
        X | qubit_r[n - 1]
        ExpMod(a, N, qubit_x, qubit_r, qubit_a,
               qubit_b, qubit_c, overflow, qubit_N, t)
        return CompositeGate(gates=circuit.gates)



import os
import json
import psutil
import shutil
import subprocess

from QuICT.tools import Logger, LogFormat


logger = Logger("Node_Job_Management", LogFormat.full)


class QuICTLocalManager:
    """ QuICT Job Management for the Local Mode. Using SQL to store running-time information. """
    def __init__(self):
        self._default_path = "/home/nodes/result"

    def _job_prepare(self, output_path: str, circuit_qasm: str):
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        with open(f"{output_path}/circuit.qasm", 'w') as cw:
            cw.write(circuit_qasm)

    def start_job(self, job_info: dict):
        """ Start the job describe by the given yaml file.

        Args:
            job_info (dict): The given job information dict
        """
        # Check job name
        name = job_info["job_name"]

        try:
            self._start_job(name, job_info)
        except Exception as e:
            logger.warn(f"Failure to start job, due to {e}.")

    def _start_job(self, job_name: str, yml_dict: dict):
        # Get information from given yml dict
        job_options = {}
        job_options["output_path"] = os.path.join(self._default_path, job_name)
        job_options["circuit_path"] = os.path.join(self._default_path, job_name, "circuit.qasm")
        self._job_prepare(job_options["output_path"], yml_dict["circuit_info"]["qasm"])

        script_name = ""
        if "simulation" in yml_dict.keys():
            script_name = "simulation"
            job_options["device"] = yml_dict["device"]
            job_options.update(yml_dict['simulation'])

        if "qcda" in yml_dict.keys():
            script_name = "qcda" if script_name == "" else "mixed_pipe"
            job_options.update(yml_dict['qcda'])

        # Pre-paration job's runtime arguments
        runtime_args = ""
        for key, value in job_options.items():
            runtime_args += f"{key}={value} "

        # Start job
        command_file_path = os.path.join(
            os.path.dirname(__file__),
            "../script",
            f"{script_name}.py"
        )
        proc = subprocess.Popen(
            f"python {command_file_path} {runtime_args}", shell=True
        )

        # Save job information into SQL DB.
        job_info = {
            'name': job_name,
            'status': 'running',
            'pid': proc.pid
        }
        with open(f"{os.path.join(self._default_path, job_name)}/job_info.json", 'w') as cw:
            cw.write(json.dumps(job_info))

        logger.info(f"Successfully start the job {job_name} in local mode.")

    def _get_job_info(self, job_name: str) -> tuple:
        """ get job info, return job_pid, job_status.

        Args:
            name (str): job's name
        """
        file_path = os.path.join(self._default_path, job_name, "job_info.json")
        job_dict = json.load(file_path)
        job_pid = job_dict["pid"]
        job_status = job_dict["status"]
        if job_status in ["finish", "stop"]:
            return job_pid, job_status

        try:
            _ = psutil.Process(job_pid)
        except:
            job_dict["status"] = "finish"
            with open(file_path, 'w') as cw:
                cw.write(json.dumps(job_dict))

        return job_pid, job_dict["status"]

    def status_job(self, job_name: str):
        """ Get job's states """
        # Check job's list contain given job
        _, status = self._get_job_info(job_name)
        return status

    def delete_job(self, job_name: str):
        """ Delete a job. """
        # check job status
        job_pid, job_status = self._get_job_info(job_name)
        # kill job's process, if job is still running
        if job_status in ["stop", "running"]:
            try:
                job_process = psutil.Process(job_pid)
                job_process.kill()
                logger.info(f"Successfully kill the job {job_name}'s process.")
            except Exception as e:
                logger.warn(f"Failure to kill the job {job_name}, due to {e}.")

        # Delete from Redis DB
        shutil.rmtree(os.path.join(self._default_path, job_name))
        logger.info(f"Successfully delete the job {job_name}.")



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2022/2/24 上午10:37
# @Author  : Kaiqi Li
# @File    : cpu_calculator
import os
import subprocess
import numpy as np

try:
    from QuICT.ops.backend import linalg
except ImportError:
    backend_file_path = os.path.dirname(__file__) + "/../backend/linear_alg_cpu.py"
    res = subprocess.call(["python3", backend_file_path])
    from QuICT.ops.backend import linalg


def MatrixTensorI(A, n, m):
    """ tensor I^n and A and I^m

    Args:
        A(np.array<np.complex>): the matrix A
        n(int): the index of indentity
        m(int): the index of indentity

    Returns:
        np.array<np.complex>: the tensor result I^n ⊗ A ⊗ I^m
    """
    # Parameter normalized
    n, m = np.int32(n), np.int32(m)
    if not isinstance(A, np.ndarray):
        A = np.array(A)

    if A.dtype == np.complex64:
        return linalg.matrixtensorf(A, n, m)
    elif A.dtype == np.complex128:
        return linalg.matrixtensord(A, n, m)
    else:
        raise TypeError(f"The matrix A should be complex64 or complex128, not {A.dtype}.")


def MatrixPermutation(A: np.ndarray, mapping: np.ndarray, changeInput: bool = False) -> np.ndarray:
    """ permute A with mapping, inplace

    Args:
        A(np.array<np.complex>): the matrix A
        mapping(np.ndarray<int>): the qubit mapping
        changeInput(bool): whether changes in A
    """
    # Parameter normalized
    assert isinstance(changeInput, bool)
    if not isinstance(A, np.ndarray):
        A = np.array(A)

    if not isinstance(mapping, np.ndarray):
        mapping = np.array(mapping, dtype=np.int32)

    if mapping.dtype != np.int32:
        mapping = mapping.astype(np.int32)

    if A.dtype == np.complex64:
        ops = linalg.matrixpermf
    elif A.dtype == np.complex128:
        ops = linalg.matrixpermd
    else:
        raise TypeError(f"The matrix A should be complex64 or complex128, not {A.dtype}.")

    if changeInput:
        ops(A, mapping, changeInput)
    else:
        return ops(A, mapping, changeInput)


def VectorPermutation(A: np.ndarray, mapping: np.ndarray, changeInput: bool = False):
    """ permutaion A with mapping, changeInput

    Args:
        A(np.array<np.complex>): the matrix A
        mapping(np.ndarray<int>): the qubit mapping
        changeInput(bool): whether changes in A
    Returns:
        np.array<np.complex>: the result of Permutation
    """
    # Parameter normalized
    assert isinstance(changeInput, bool)
    if not isinstance(A, np.ndarray):
        A = np.array(A)

    if not isinstance(mapping, np.ndarray):
        mapping = np.array(mapping, dtype=np.int32)

    if mapping.dtype != np.int32:
        mapping = mapping.astype(np.int32)

    if A.dtype == np.complex64:
        ops = linalg.vectorpermf
    elif A.dtype == np.complex128:
        ops = linalg.vectorpermd
    else:
        raise TypeError(f"The matrix A should be complex64 or complex128, not {A.dtype}.")

    if changeInput:
        ops(A, mapping, changeInput)
    else:
        return ops(A, mapping, changeInput)


def tensor(A, B):
    """ tensor A and B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B

    Returns:
        np.array<np.complex>: the tensor result A ⊗ B
    """
    # Parameter normalized
    if not isinstance(A, np.ndarray):
        A = np.array(A)

    if not isinstance(B, np.ndarray):
        B = np.array(B)

    if A.dtype == np.complex64 and B.dtype == np.complex64:
        return linalg.tensorf(A, B)
    elif A.dtype == np.complex128 and B.dtype == np.complex128:
        return linalg.tensord(A, B)
    else:
        raise TypeError(f"The matrix should be complex64 or complex128, not {A.dtype} and {B.dtype}.")


def dot(A, B):
    """ dot matrix A and matrix B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B

    Returns:
        np.array<np.complex>: A * B
    """
    # Parameter normalized
    if not isinstance(A, np.ndarray):
        A = np.array(A)

    if not isinstance(B, np.ndarray):
        B = np.array(B)

    assert A.dtype == B.dtype and A.shape[1] == B.shape[0]
    if A.dtype == np.complex64:
        if B.ndim == 1:
            return linalg.vdotmf(A, B)
        else:
            return linalg.dotf(A, B)
    elif A.dtype == np.complex128:
        if B.ndim == 1:
            return linalg.vdotmd(A, B)
        else:
            return linalg.dotd(A, B)
    else:
        raise TypeError(f"The matrix should be complex64 or complex128, not {A.dtype} and {B.dtype}.")


def multiply(A, B):
    """ multiply matrix A and matrix B

    Args:
        A(np.array<np.complex>): the matrix A
        B(np.array<np.complex>): the matrix B

    Returns:
        np.array<np.complex>: A x B
    """
    # Parameter normalized
    if not isinstance(A, np.ndarray):
        A = np.array(A)

    if not isinstance(B, np.ndarray):
        B = np.array(B)

    if A.dtype == np.complex64 and B.dtype == np.complex64:
        return linalg.multiplyf(A, B)
    elif A.dtype == np.complex128 and B.dtype == np.complex128:
        return linalg.multiplyd(A, B)
    else:
        raise TypeError(f"The matrix should be complex64 or complex128, not {A.dtype} and {B.dtype}.")


def matrix_dot_vector(
    vec: np.ndarray,
    vec_bit: int,
    mat: np.ndarray,
    mat_bit: int,
    mat_args: np.ndarray
):
    """ Dot the quantum gate's matrix and qubits'state vector, depending on the target qubits of gate.

    Args:
        mat (np.ndarray): The 2D numpy array, represent the quantum gate's matrix
        mat_bit (np.int32): The quantum gate's qubit number
        vec (np.ndarray): The state vector of qubits
        vec_bit (np.int32): The number of qubits
        mat_args (np.ndarray): The target qubits of quantum gate

    Raises:
        TypeError: matrix and vector should be complex and with same precision

    Returns:
        np.ndarray: updated state vector
    """
    # Parameter normalized
    mat_bit, vec_bit = np.int32(mat_bit), np.int32(vec_bit)
    if not isinstance(mat, np.ndarray):
        mat = np.array(mat)

    if not isinstance(vec, np.ndarray):
        vec = np.array(vec)

    if not isinstance(mat_args, np.ndarray):
        mat_args = np.array(mat_args, np.int32)
    elif mat_args.dtype != np.int32:
        mat_args = mat_args.astype(np.int32)

    if mat_bit == vec_bit:
        return dot(mat, vec)

    if mat.dtype == np.complex64 and vec.dtype == np.complex64:
        return linalg.matdotvecf(mat, mat_bit, vec, vec_bit, mat_args)
    elif mat.dtype == np.complex128 and vec.dtype == np.complex128:
        return linalg.matdotvecd(mat, mat_bit, vec, vec_bit, mat_args)
    else:
        raise TypeError(f"The matrix should be complex64 or complex128, not {mat.dtype} and {vec.dtype}.")


def measure_gate_apply(
    index: int,
    vec: np.array
):
    """ Measured the state vector for target qubit

    Args:
        index (int): The index of target qubit
        vec (np.array): The state vector of qubits

    Raises:
        TypeError: The vector should be complex.

    Returns:
        _type_: The updated state vector
    """
    # Parameter normalized
    index = np.int32(index)
    if not isinstance(vec, np.ndarray):
        vec = np.array(vec)

    if vec.dtype == np.complex64:
        return linalg.measuref(index, vec)
    elif vec.dtype == np.complex128:
        return linalg.measured(index, vec)
    else:
        raise TypeError(f"The vector should be complex64 or complex128, not {vec.dtype}.")



from __future__ import annotations

import copy
import random
from math import factorial, log, sqrt
from typing import Dict, List, Optional, Tuple

from QuICT.core import *
from QuICT.core.gate.composite_gate import CompositeGate

from ..common import CircuitInfo
from ..common.layout_info import LayoutInfo


class MCTSTreeNode:
    def __init__(
        self,
        circuit_info: CircuitInfo,
        layout_info: LayoutInfo,
        parent: Optional[MCTSTreeNode] = None,
        logic2phy: Optional[List[int]] = None,
        phy2logic: Optional[List[int]] = None,
    ) -> None:
        self.circ_info = circuit_info
        self.layout_info = layout_info
        self.executed_circ = CompositeGate()
        self.children: Dict[Tuple[int, int], MCTSTreeNode] = {}
        self.q_value: float = 0.0
        self.parent = parent
        self.transition_reward = 0.0
        self.visit_cnt: int = 1

        if logic2phy is None:
            logic2phy = list(range(self.circ_info._qubit_num))
        self.logic2phy = logic2phy

        if phy2logic is None:
            phy2logic = [-1 for _ in range(self.circ_info._qubit_num)]
            for idx, val in enumerate(self.logic2phy):
                phy2logic[val] = idx
        self.phy2logic = phy2logic

    def _update_mapping(self, action: Tuple[int, int]):
        u, v = action
        self.phy2logic[u], self.phy2logic[v] = self.phy2logic[v], self.phy2logic[u]
        lu, lv = self.phy2logic[u], self.phy2logic[v]
        self.logic2phy[lu], self.logic2phy[lv] = self.logic2phy[lv], self.logic2phy[lu]

    def _recommended_action(self) -> List[Tuple[int, int]]:
        assert not self.is_terminated_node()
        answer = []
        relative_qubit = set()
        for gate in self.circ_info.first_layer_gates.values():
            if gate.controls + gate.targets == 1:
                continue
            a, b = gate.cargs + gate.targs
            _a, _b = self.logic2phy[a], self.logic2phy[b]
            relative_qubit.add(_a)
            relative_qubit.add(_b)
        for a, b in self.layout_info.topo_edges:
            if a in relative_qubit or b in relative_qubit:
                answer.append((a, b))
        assert len(answer) > 0
        return answer

    def _expand_one(self, action: Tuple[int, int]) -> MCTSTreeNode:
        successor = MCTSTreeNode(
            circuit_info=copy.deepcopy(self.circ_info),
            layout_info=self.layout_info,
            parent=self,
            logic2phy=copy.copy(self.logic2phy),
            phy2logic=copy.copy(self.phy2logic),
        )
        successor._update_mapping(action=action)
        cg = CompositeGate()
        reward = successor.circ_info.eager_exec(
            logic2phy=successor.logic2phy,
            topo_graph=successor.layout_info.topo_graph,
            physical_circ=cg,
        )
        successor.executed_circ = cg
        successor.transition_reward = reward
        return successor

    def select_child(self, c: float) -> MCTSTreeNode:
        self.visit_cnt += 1
        candidates = list(self.children.values())
        candidates.sort(
            key=lambda x: x.q_value
            + x.transition_reward
            + c * sqrt(log(self.visit_cnt) / x.visit_cnt),
            reverse=True,
        )
        return candidates[0]

    def expand(self):
        assert len(self.children) == 0, "Only leaf node can be expanded!"
        for action in self._recommended_action():
            child = self._expand_one(action=action)
            self.children[action] = child
        assert len(self.children) > 0, "After expansion, node must have successors!"

    def _biased_random_simulate(self, sim_gate_num: int, epsilon: float) -> float:
        if self.is_terminated_node():
            return 0.0
        remove_cnt = 0
        cur = MCTSTreeNode(
            circuit_info=copy.deepcopy(self.circ_info),
            layout_info=self.layout_info,
            parent=self.parent,
            logic2phy=copy.copy(self.logic2phy),
            phy2logic=copy.copy(self.phy2logic),
        )
        step = 0
        while remove_cnt < sim_gate_num and not cur.is_terminated_node():
            candidates = cur._recommended_action()
            impact_factors = []
            for u, v in candidates:
                cur_d, nxt_d = 0, 0
                for gate in cur.circ_info.first_layer_gates.values():
                    a, b = gate.cargs + gate.targs
                    _a, _b = cur.logic2phy[a], cur.logic2phy[b]
                    cur_d += cur.layout_info.topo_dist[_a][_b]
                    l2p = copy.copy(cur.logic2phy)
                    lu, lv = cur.phy2logic[u], cur.phy2logic[v]
                    l2p[lu], l2p[lv] = l2p[lv], l2p[lu]
                    _a, _b = l2p[a], l2p[b]
                    nxt_d += cur.layout_info.topo_dist[_a][_b]
                factor = cur_d - nxt_d
                if abs(factor) < 1e-6:
                    factor += epsilon
                elif factor < 0:
                    factor = 0
                impact_factors.append(factor)
            action = random.choices(
                population=candidates,
                weights=impact_factors,
                k=1,
            )[0]
            suc = cur._expand_one(action)
            remove_cnt += suc.transition_reward
            del cur
            cur = suc
            step += 1
        return float(step)

    def simulate(self, sim_cnt: int, sim_gate_num: int, gamma: float, epsilon: float):
        assert len(self.children) == 0, "Only leaf node can be simulated!"
        # Set a large initial value
        best = float(
            10 * self.circ_info.count_gate() * self.layout_info.layout.qubit_number
        )
        for _ in range(sim_cnt):
            cur = self._biased_random_simulate(sim_gate_num, epsilon)
            best = min(best, cur)
        sim_res = gamma ** (best / 2) * float(sim_gate_num)
        self.q_value = sim_res

    def back_propagate(self, gamma: float):
        parent = self.parent
        cur = self
        while parent is not None:
            parent.q_value = max(
                parent.q_value, gamma * (self.transition_reward + cur.q_value)
            )
            cur = parent
            parent = parent.parent

    def best_child(self) -> Tuple[Tuple[int, int], MCTSTreeNode, bool]:
        """Best child's action, state, and terminate info.

        Returns:
            Tuple[Tuple[int, int], MCTSTreeNode, bool]:
                Action to child, child node, whether mapping ends.
        """
        candidates = list(self.children.items())
        action, child = max(
            candidates, key=lambda x: x[1].q_value + x[1].transition_reward
        )
        terminated = child.is_terminated_node()
        return action, child, terminated

    def is_terminated_node(self) -> bool:
        return self.circ_info.count_gate() == 0



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2021/1/11 12:26 下午
# @Author  : Han Yu
# @File    : cnot_bfs.py

import os

import numpy as np

from QuICT.qcda.optimization.cnot_local_force.utility.utility import apply_cx, count, path


def generate_json(n):
    """ find the best circuit by bfs

    Args:
        n(int): input qubits
    """
    now = 0
    for i in range(n):
        now ^= (1 << (i * n + i))
    vis = np.zeros(1 << (n * n), dtype=int)
    vis[now] = 1
    pre = [None] * (1 << (n * n))
    out = {}
    queue = [now]
    out[now] = []
    l = 0
    ans = count(n)
    total = 1
    while True:
        now = queue[l]
        for i in range(n):
            for j in range(n):
                if i != j:
                    new_state = apply_cx(now, i, j, n)
                    if vis[new_state] == 0:
                        pre[new_state] = path(now, i, j)
                        vis[new_state] = vis[now] + 1
                        queue.append(new_state)
                        paths = []
                        total += 1
                        while new_state != queue[0]:
                            paths.append(pre[new_state].CX_tuple)
                            new_state = pre[new_state].father_node
                        out[queue[-1]] = []
                        for index in range(len(paths) - 1, -1, -1):
                            out[queue[-1]].append((paths[index]))
        if total == ans:
            break
        l += 1
    # json_data = json.dumps(out)
    # with open("./bfs/" + str(n) + 'qubit_cnot.json', 'w+') as file:
    #    file.write(json_data)
    keys = out.keys()
    with open(f"{os.path.dirname(os.path.abspath(__file__))}{os.path.sep}{n}qubit_cnot.inf", 'w') as file:
        file.write(";")
        for key in keys:
            string = f"{key}:"
            tuples = out[key]
            len_tuples = len(tuples)
            for i in range(len_tuples):
                _tuple = tuples[i]
                string += f"{_tuple[0] * 5 + _tuple[1]}"
                if i + 1 < len_tuples:
                    string += ','
                else:
                    string += ';'
            if len_tuples == 0:
                string += ';'
            file.write(string)



"""
Optimize the given Circuit/CompositeGate by merging the adjacent gates with
the commutative relation between gates in consideration.
"""

import numpy as np

from QuICT.core.gate import *
from QuICT.qcda.utility import OutputAligner

# Categories of combination
elimination = [
    GateType.h, GateType.x, GateType.y, GateType.z, GateType.cx, GateType.hy,
    GateType.cy, GateType.cz, GateType.ch, GateType.ccx, GateType.swap
]
addition = [
    GateType.u1, GateType.rx, GateType.ry, GateType.rz, GateType.phase, GateType.gphase,
    GateType.crz, GateType.cu1, GateType.fsim, GateType.rxx, GateType.ryy, GateType.rzz
]
multiplication = [GateType.unitary]
other = [GateType.sx, GateType.sy, GateType.s, GateType.sdg, GateType.t, GateType.tdg]
not_calculated = [GateType.sw, GateType.u2, GateType.u3, GateType.cu3]


class Node(object):
    """
    (Temporary) implementation of Directed Acyclic Graph (DAG) used in this code

    TODO: Replace this part with a graph structure
    """
    def __init__(self, gate: BasicGate):
        """
        Args:
            gate(BasicGate): Gate represented by the node
            identity(bool): Whether the gate is identity (upon a global phase)
            predecessor(list[int]): Predecessors of the node
            reachable(bool): Whether this node needs to be compared with the new node
        """
        self.gate = gate
        self.identity = False
        self.predecessor = set()
        self.reachable = True


class CommutativeOptimization(object):
    """
    Optimize the given Circuit/CompositeGate by merging the adjacent gates with
    the commutative relation between gates in consideration.
    """
    para_rule = {
        GateType.x: (Rx(np.pi), np.pi / 2),
        GateType.sx: (Rx(np.pi / 2), np.pi / 4),
        GateType.y: (Ry(np.pi), np.pi / 2),
        GateType.sy: (Ry(np.pi / 2), 0),
        GateType.z: (Rz(np.pi), np.pi / 2),
        GateType.s: (Rz(np.pi / 2), np.pi / 4),
        GateType.sdg: (Rz(-np.pi / 2), -np.pi / 4),
        GateType.t: (Rz(np.pi / 4), np.pi / 8),
        GateType.tdg: (Rz(-np.pi / 4), -np.pi / 8)
    }

    depara_rule = {
        # Rx
        (GateType.rx, 0): ([ID], 0),
        (GateType.rx, 2): ([SX], -np.pi / 4),
        (GateType.rx, 4): ([X], -np.pi / 2),
        (GateType.rx, 6): ([X, SX], -3 * np.pi / 4),
        (GateType.rx, 8): ([ID], np.pi),
        (GateType.rx, 10): ([SX], 3 * np.pi / 4),
        (GateType.rx, 12): ([X], np.pi / 2),
        (GateType.rx, 14): ([X, SX], np.pi / 4),
        # Ry
        (GateType.ry, 0): ([ID], 0),
        (GateType.ry, 2): ([SY], 0),
        (GateType.ry, 4): ([Y], -np.pi / 2),
        (GateType.ry, 6): ([Y, SY], -np.pi / 2),
        (GateType.ry, 8): ([ID], np.pi),
        (GateType.ry, 10): ([SY], np.pi),
        (GateType.ry, 12): ([Y], np.pi / 2),
        (GateType.ry, 14): ([Y, SY], np.pi / 2),
        # Rz
        (GateType.rz, 0): ([ID], 0),
        (GateType.rz, 1): ([T], -np.pi / 8),
        (GateType.rz, 2): ([S], -np.pi / 4),
        (GateType.rz, 3): ([S, T], -3 * np.pi / 8),
        (GateType.rz, 4): ([Z], -np.pi / 2),
        (GateType.rz, 5): ([Z, T], -5 * np.pi / 8),
        (GateType.rz, 6): ([S_dagger], 5 * np.pi / 4),
        (GateType.rz, 7): ([T_dagger], 9 * np.pi / 8),
        (GateType.rz, 8): ([ID], np.pi),
        (GateType.rz, 9): ([T], 7 * np.pi / 8),
        (GateType.rz, 10): ([S], 3 * np.pi / 4),
        (GateType.rz, 11): ([S, T], 5 * np.pi / 8),
        (GateType.rz, 12): ([Z], np.pi / 2),
        (GateType.rz, 13): ([Z, T], 3 * np.pi / 8),
        (GateType.rz, 14): ([S_dagger], np.pi / 4),
        (GateType.rz, 15): ([T_dagger], np.pi / 8)
    }

    def __init__(self, parameterization=True, deparameterization=False):
        """
        Args:
            parameterization(bool, optional): whether to use the parameterize() process
            deparameterization(bool, optional): whether to use the deparameterize() process
        """
        self.parameterization = parameterization
        self.deparameterization = deparameterization

    def __repr__(self):
        return f'CommutativeOptimization(parameterization={self.parameterization}, ' \
               f'deparameterization={self.deparameterization})'

    @classmethod
    def parameterize(cls, gate: BasicGate):
        """
        In BasicGates, (X, SX), (Y, SY), (Z, S, Sdagger, T, Tdagger) could be
        'parameterized' to Rx, Ry, Rz respectively, which is helpful in the
        `combine` function.

        Args:
            gate(BasicGate): Gate to be transformed to its 'parameterized' version

        Returns:
            Tuple[BasicGate, float]: If the `gate` is listed above, its 'parameterized'
                version with the phase angle derived in the process will be returned.
                Otherwise, the `gate` itself with phase angle 0 will be returned.
        """
        try:
            gate_para, phase = cls.para_rule[gate.type]
            return gate_para & gate.targ, phase
        except KeyError:
            return gate, 0

    @classmethod
    def deparameterize(cls, gate: BasicGate):
        """
        Deparameterize the parameterized gates if possible, as an inverse process of
        `parameterize` function.

        Be aware that gates like Rx(3*np.pi/2) would be transformed to X.SX (which would cause more gates).

        Args:
            gate(BasicGate): Gate to be transformed to its 'deparameterized' version

        Returns:
            Tuple[CompositeGate, float]: If deparameterization process is possible, the
                'deparameterized' version of the gate with the phase angle derived in
                the process will be returned. Otherwise, the `gate` itself with phase
                angle 0 will be returned.
        """
        gates_depara = CompositeGate()
        try:
            parg = np.mod(gate.parg, 4 * np.pi) / (np.pi / 4)
            assert np.isclose(round(parg), parg)
            g_list, phase = cls.depara_rule[gate.type, round(parg)]
            for g in g_list:
                gates_depara.append(g & gate.targ)
            return gates_depara, phase
        except Exception:
            gates_depara.append(gate)
            return gates_depara, 0

    @staticmethod
    def combine(gate_x: BasicGate, gate_y: BasicGate):
        """
        Combine `gate_x` and `gate_y` of the same type

        Generally, the combination could be divided into four categories:
        1. Elimination: the combined gate is ID
        2. Addition: the parameters of gates should be added
        3. Multiplication: the matrices of gates should be multiplied(i.e. UnitaryGate)
        4. Other: some special case(e.g. SS=Z) or not able to be calculated easily(e.g. U3Gate)
        In this method we would only deal with the first 3 cases, while the last case is partially
        handled by preprocessing the `parameterize` function.

        Args:
            gate_x(BasicGate): Gate to be combined
            gate_y(BasicGate): Gate to be combined

        Returns:
            BasicGate: The combined gate

        Raises:
            TypeError: If the input gates are not of the same type or unknown gate type encountered.
            ValueError: If the input gates are not operating on the same qubits in the same way
                or could not be combined directly to a gate with the same type.
        """
        assert gate_x.type == gate_y.type,\
            TypeError('commu_opt', 'Gates with same type', 'different type.')
        assert gate_x.cargs == gate_y.cargs and gate_x.targs == gate_y.targs,\
            ValueError('commu_opt', 'same control and target qubits', 'different qubits.')

        if gate_x.type in elimination:
            # IDGates operating on all qubits are the same
            return ID.copy() & gate_x.targ

        if gate_x.type in addition:
            gate = gate_x.copy()
            for id_para in range(gate_x.params):
                if gate_x.type in [GateType.u1, GateType.cu1, GateType.fsim]:
                    gate.pargs[id_para] = np.mod(gate_x.pargs[id_para] + gate_y.pargs[id_para], 2 * np.pi)
                else:
                    gate.pargs[id_para] = np.mod(gate_x.pargs[id_para] + gate_y.pargs[id_para], 4 * np.pi)
            return gate

        if gate_x.type in multiplication:
            gate = gate_x.copy()
            gate.matrix = gate_y.matrix.dot(gate_x.matrix)
            return gate

        if gate_x.type in other or gate_x.type in not_calculated:
            raise ValueError('commu_opt', 'calculated', 'not')

        raise TypeError('commu_opt', 'BasicGate', f'{gate_x.type}')

    @OutputAligner()
    def execute(self, gates):
        """
        Optimize the given Circuit/CompositeGate by merging the adjacent gates with
        the commutative relation between gates in consideration.

        WARNING: This method is implemented for Circuit/CompositeGate with BasicGates
        only (say, ComplexGates are not supported), other gates in the Circuit/
        CompositeGate may result in an exception or unexpected output.

        FIXME: Merging gates may cause the modification of commutative relation.
        In this version only the simplest (also the most common) case, i.e. the merged
        gate is identity, is handled. More specified analysis of the DAG is needed
        to deal with other cases, which is postponed until the graph structure is completed.

        Args:
            gates(Circuit/CompositeGate): Circuit/CompositeGate to be optimized

        Returns:
            CompositeGate: The CompositeGate after optimization
        """
        gates = CompositeGate(gates=gates.gates)
        nodes: list[Node] = []
        phase_angle = 0

        # Greedy optimizationn
        for gate in gates:
            gate: BasicGate
            # IDGate
            if gate.type == GateType.id:
                continue

            # GlobalPhaseGate
            if gate.type == GateType.gphase:
                phase_angle += gate.parg
                continue

            # Remove such as Rot(0)
            if np.allclose(
                gate.matrix,
                gate.matrix[0, 0] * np.eye(1 << gate.controls + gate.targets)
            ):
                phase_angle += np.angle(gate.matrix[0, 0])
                continue

            # Preprocess: parameterization
            if self.parameterization:
                gate, phase = self.parameterize(gate)
                phase_angle += phase
            new_node = Node(gate)

            # Main Procedure
            length = len(nodes)
            for prev in range(length):
                if nodes[prev].identity:
                    nodes[prev].reachable = False
                else:
                    nodes[prev].reachable = True

            combined = False
            for prev in range(length - 1, -1, -1):
                prev_node = nodes[prev]
                if prev_node.reachable:
                    prev_gate = prev_node.gate
                    # Combination of prev_gate and gate if same type
                    if (
                        prev_gate.type == gate.type and
                        prev_gate.cargs == gate.cargs and
                        prev_gate.targs == gate.targs and
                        not (gate.type in not_calculated)
                    ):
                        combined = True
                        nodes[prev].gate = self.combine(prev_gate, gate)
                        mat = nodes[prev].gate.matrix
                        if (
                            nodes[prev].gate.type == GateType.id or
                            np.allclose(
                                mat,
                                mat[0, 0] * np.eye(1 << nodes[prev].gate.controls + nodes[prev].gate.targets)
                            )
                        ):
                            nodes[prev].identity = True
                        break

                    if not prev_gate.commutative(gate):
                        for node in prev_node.predecessor:
                            nodes[node].reachable = False
                        new_node.predecessor.add(prev)
                        new_node.predecessor = new_node.predecessor.union(prev_node.predecessor)

            if not combined:
                nodes.append(new_node)

        gates_opt = CompositeGate()
        for node in nodes:
            if node.identity or node.gate.type == GateType.gphase:
                phase_angle += np.angle(node.gate.matrix[0, 0])
            elif self.deparameterization:
                gates_depara, phase = self.deparameterize(node.gate)
                gates_opt.extend(gates_depara)
                phase_angle += phase
            else:
                gates_opt.append(node.gate)

        phase_angle = np.mod(phase_angle.real, 2 * np.pi)
        if not np.isclose(phase_angle, 0) and not np.isclose(phase_angle, 2 * np.pi):
            with gates_opt:
                GPhase(phase_angle) & 0

        return gates_opt



#!/usr/bin/env python
# -*- coding:utf8 -*-
# @TIME    : 2020/8/22 2:40
# @Author  : Han Yu
# @File    : __init__.py

from .arithmetic import *
from .clifford import *
from .gate_transform import *
from .quantum_state_preparation import *
from .unitary_decomposition import *



from collections import Iterable, deque
from typing import Set

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np

from QuICT.core.gate import BasicGate
from QuICT.tools.exception.core import TypeError


class DAGNode:
    """ The node in DAG Circuit, represent a quantum gate in circuit.

    Args:
        id (int): The unique identity, usually be the quantum gates' index
        gate (BasicGate): The quantum gate
        successors (list, optional): The successors for nodes. Defaults to [].
        predecessors (list, optional): The predecessors for nodes. Defaults to [].
    """
    @property
    def id(self):
        return self._id

    @property
    def gate(self):
        return self._gate

    @property
    def name(self):
        return self._name

    @property
    def cargs(self):
        return self._cargs

    @property
    def targs(self):
        return self._targs

    @property
    def qargs(self):
        return self._qargs

    @property
    def successors(self):
        return self._successors

    @property
    def type(self):
        return self._type

    @successors.setter
    def successors(self, sces: list):
        self._successors = sces

    @property
    def predecessors(self):
        return self._predecessors

    @predecessors.setter
    def predecessors(self, pdces: list):
        self._predecessors = pdces

    def __init__(self, id: int, gate: BasicGate, successors: list = None, predecessors: list = None):
        self._id = id
        self._gate = gate
        self._name = gate.qasm_name
        self._cargs = gate.cargs
        self._targs = gate.targs
        self._qargs = gate.cargs + gate.targs
        self._type = gate.type
        self._successors = [] if successors is None else successors
        self._predecessors = [] if predecessors is None else predecessors

    def append_pred(self, u):
        """
        Append a node id to successors.

        Args:
            u(int): the node
        """
        self._predecessors.append(u)

    def append_succ(self, u):
        """
        Append a node id to predecessors.

        Args:
            u(int): the node
        """

        self._successors.append(u)


class DAGCircuit:
    """ The DAG Circuit using networkx.DiGraph()

    The nodes in the graph represented the quantum gates, and the edges means the two quantum
    gates is non-commutation. In other words, a directed edge between node A with quantum gate GA
    and node B with quantum gate GB, the quantum gate GA does not commute with GB.

    The nodes in the graph have the following attributes:
    'name', 'gate', 'cargs', 'targs', 'qargs', 'successors', 'predecessors'.

    **Reference:**

    [1] Iten, R., Moyard, R., Metger, T., Sutter, D. and Woerner, S., 2020.
    Exact and practical pattern matching for quantum circuit optimization.
    `arXiv:1909.05270 <https://arxiv.org/abs/1909.05270>`_

    Args:
        circuit (Circuit): The quantum circuit.
    """
    @property
    def name(self) -> str:
        return self._name

    @property
    def size(self) -> int:
        return self._size

    @property
    def width(self) -> int:
        return self._width

    @property
    def gates(self):
        for node_id in self.nodes():
            node = self.get_node(node_id)
            yield node.gate

    def __init__(self, circuit, node_type=DAGNode):
        self._circuit = circuit
        self._name = f"DAG_{self._circuit.name}"
        self._size = self._circuit.size()
        self._width = self._circuit.width()
        self._graph = nx.DiGraph()
        self._node_type = node_type
        # Build DAG Circuit
        self._to_dag_circuit()

    ####################################################################
    ############         DAGCircuit Nodes and Edges         ############
    ####################################################################
    def __getitem__(self, item):
        """ to fit the slice operator, overloaded this function.

        get a smaller qureg/qubit from this circuit

        Args:
            item(int/slice): slice passed in.
        Return:
            Qubit/Qureg: the result or slice
        """
        return self.nodes()[item]

    def nodes(self) -> list:
        """ Get all nodes in DAG.

        Returns:
            list: The list of nodes
        """
        return list(self._graph.nodes)

    def add_node(self, node: DAGNode):
        """ Add a node into DAG.

        Args:
            node (DAGNode): The DAG Node
        """
        assert isinstance(node, DAGNode), TypeError("DagCircuit.add_node.node", "DAGNode", type(node))
        if not self._graph.has_node(node.id):
            self._graph.add_node(node.id, node=node)

    def get_node(self, node_id: int):
        """ Get DAG node's data.

        Args:
            node_id (int): the unique identity for node
        """
        return self._graph.nodes[node_id]["node"]

    def edges(self) -> list:
        """ Get all edges in DAG.

        Returns:
            list: The list of edges of DAG.
        """
        return list(self._graph.edges)

    def in_edges(self, node_id: int):
        """ Get all in-edges of node with given node id.

        Args:
            node_id (int): the unique identity for node
        """
        return self._graph.in_edges(node_id)

    def out_edges(self, node_id: int):
        """ Get all out-edges of node with given node id.

        Args:
            node_id (int): the unique identity for node
        """
        return self._graph.out_edges(node_id)

    ####################################################################
    ############                Circuit to DAG              ############
    ####################################################################
    def _to_dag_circuit(self):
        gates = self._circuit.gates
        reachable = np.zeros(shape=(len(gates), ), dtype=bool)
        for idx, g in enumerate(gates):
            assert isinstance(g, BasicGate), "Only support BasicGate in DAGCircuit."
            cur = self._node_type(idx, g)
            self.add_node(cur)
            reachable[: idx] = True
            for prev in reversed(range(idx)):
                if reachable[prev] and not g.commutative(gates[prev]):
                    self.add_edge(prev, idx)
                    reachable[list(self.all_predecessors(prev))] = False

    def _all_reachable(self, start, direction):
        if isinstance(start, int):
            visited = {start}
        elif isinstance(start, Iterable):
            visited = set(start)
        else:
            raise TypeError("DagCircuit.all_successors/predecessors.start", "int/Iterable", type(start))

        que = deque(visited)
        init_visited = visited.copy()
        while len(que) > 0:
            cur_node = self.get_node(que.popleft())
            for node_id in getattr(cur_node, direction):
                if node_id not in visited:
                    que.append(node_id)
                    visited.add(node_id)

        return visited - init_visited

    def all_predecessors(self, start) -> Set[int]:
        """
        Return all predecessors (direct and indirect) of `start`.
        `start` can be a node id (int) or many node ids (Iterable).

        Args:
            start(int/Iterable): the start node(s)

        Returns:
            set: set of predecessors
        """

        return self._all_reachable(start, 'predecessors')

    def all_successors(self, start) -> Set[int]:
        """
        Return all successors (direct and indirect) of `start`.

        Args:
            start(int/Iterable): the start node(s)

        Returns:
            Set[int]: set of successors
        """

        return self._all_reachable(start, 'successors')

    def add_edge(self, u, v):
        """
        Add a directed edge to DAG.

        Args:
            u(int): start node
            v(int): end node
        """

        self._graph.add_edge(u, v)
        self.get_node(u).append_succ(v)
        self.get_node(v).append_pred(u)

    ####################################################################
    ############              DAG_Circuit utils             ############
    ####################################################################
    def draw(self, layout=nx.shell_layout):
        """ Draw a DAG circuit, save as jpg file.

        Args:
            layout (layout, optional): The networkx.layout. Defaults to nx.shell_layout.
        """
        graph_name = f"{self.name}.jpg"
        plt.figure()
        nx.draw(self._graph, pos=layout(self._graph), with_labels=True)
        plt.savefig(graph_name)



from QuICT.core.gate import CompositeGate, CX, CH, X, H
from QuICT.core.gate.backend import MCTOneAux


class OracleInfo:
    def __init__(
        self,
        n: int,
        n_ancilla: int,
        S_chi,
        is_good_state,
        custom_grover_operator=None,
    ) -> None:
        """Oracle information.

        Args:
            n (int): oracle space width.
            n_ancilla (int, optional): ancilla width.
            S_chi (function, optional): a function,\
                input of which is (n:int),\
                output of which is the target phase-flip circuit of width n.
            is_good_state (function, optional): a function,\
                input of which is (s:String),\
                output of which is whether the bit string s is a good state.
            custom_grover_operator (CompositeGate, optional): user-defined Grover operator. Defaults to None.
        """
        self.n = n
        self.n_ancilla = n_ancilla
        self.S_chi = S_chi
        self.is_good_state = is_good_state
        self.custom_grover_operator = custom_grover_operator


def default_A(n, controlled=False):
    # control on 0
    cgate = CompositeGate()
    if controlled:
        for i in range(1, 1 + n):
            CH | cgate([0, i])
    else:
        for i in range(n):
            H | cgate(i)
    return cgate


def default_A_dagger(n, controlled=False):
    return default_A(n, controlled)


def default_S_0(n, controlled=False):
    # control on 0
    cgate = CompositeGate()
    if controlled:
        indices = list(range(1, 1 + n))
        for i in indices:
            CX | cgate([0, i])
        CH | cgate([0, indices[n - 1]])
        MCTOneAux().execute(n + 2) | cgate
        CH | cgate([0, indices[n - 1]])
        for i in indices:
            CX | cgate([0, i])
    else:
        for i in range(n):
            X | cgate(i)
        H | cgate(n - 1)
        MCTOneAux().execute(n + 1) | cgate
        H | cgate(n - 1)
        for i in range(n):
            X | cgate(i)
    return cgate


class StatePreparationInfo:
    def __init__(self, n: int, n_ancilla=0, A=default_A, A_dagger=default_A_dagger) -> None:
        """state preparation information.

        Args:
            n (int): state preparation circuit width.
            n_ancilla (int, optional): ancilla width. Defaults to 0.
            A (function, optional): a function that has same input/output as default_A. Defaults to default_A.
            A_dagger (function, optional): a function that has same input/output as default_A_dagger,\
                and it should be consistent with A. Defaults to default_A_dagger.
        """
        self.n = n
        self.n_ancilla = n_ancilla
        self.A = A
        self.A_dagger = A_dagger
        self.S_0 = default_S_0



from collections import deque
from collections.abc import Iterable
from itertools import chain
from typing import Dict, List, Set, Tuple

from QuICT.core import *
from QuICT.core.gate import *
from QuICT.core.gate.gate_builder import GATE_TYPE_TO_CLASS

from .symbolic_phase import SymbolicPhase, SymbolicPhaseVariable


class DAG(Iterable):
    """
    DAG representation of a quantum circuit that indicates the commutative
    relations between gates. Iterate over a DAG will gate a sequence of
    BasicGate's in topological order.
    """

    class Node:
        """
        DAG node class.
        """

        __slots__ = ['predecessors', 'successors', 'flag', 'qubit_id', 'size', 'qubit_loc', 'qubit_flag',
                     'gate_type', '_params', 'poly_phase']

        FLAG_DEFAULT = 0
        FLAG_VISITED = 1
        FLAG_ERASED = -1
        FLAG_IN_QUE = 2
        FLAG_TO_ERASE = 3

        def __init__(self, gate_: BasicGate = None, qubit_=0):
            """
            Args:
                gate_(BasicGate): Gate represented by this node
                qubit_(int): the actual qubit the gate sits on (used only when `gate_` is None)
            """
            self.gate_type = gate_.type if gate_ is not None else None
            self._params = gate_.pargs.copy() if gate_ is not None else []

            # the actual qubit the i-th wire of the gate act on
            self.qubit_loc = [qubit_] if gate_ is None else list(chain(gate_.cargs, gate_.targs))
            # inverse mapping of self.qubit_loc
            self.qubit_id = {qubit_: 0} if gate_ is None else {qubit_: i for i, qubit_ in enumerate(self.qubit_loc)}

            self.size = len(self.qubit_id)
            self.predecessors: List[Tuple[DAG.Node, int]] = [(None, 0)] * self.size
            self.successors: List[Tuple[DAG.Node, int]] = [(None, 0)] * self.size

            # temp variables used for optimization algorithms
            self.qubit_flag = [self.FLAG_DEFAULT] * self.size
            self.flag = self.FLAG_DEFAULT
            self.poly_phase = None

        @property
        def params(self):
            return self._params

        @params.setter
        def params(self, args):
            assert type(args) == list, 'assignment of params must be a list'
            self._params = args

        def get_gate(self):
            """
            Get a copy of corresponding gate of the node

            Returns:
                BasicGate: corresponding gate
            """

            return GATE_TYPE_TO_CLASS[self.gate_type]()(*self.params) & self.qubit_loc \
                if self.gate_type is not None else None

        def add_forward_edge(self, qubit_, node):
            """
            Add an edge from self to node. Make sure qubit_id is up to date.

            Args:
                qubit_(int): the actual qubit this edge sit on
                node(DAG.Node): the node this edge point to
            """
            u_id = self.qubit_id[qubit_]
            v_id = node.qubit_id[qubit_]
            self.successors[u_id] = (node, v_id)
            node.predecessors[v_id] = (self, u_id)

        def connect(self, forward_qubit, backward_qubit, node):
            """
            Connect (self, forward_qubit) to (node, backward_qubit)

            Args:
                forward_qubit(int): the wire this edge points from
                backward_qubit(int): the wire this edge points to
                node(DAG.Node):  the node this edge points to
            """
            self.successors[forward_qubit] = (node, backward_qubit)
            node.predecessors[backward_qubit] = (self, forward_qubit)

        def append(self, forward_qubit, backward_qubit, node):
            """
            Insert a (node, backward_qubit) after (self, forward_qubit)

            Args:
                forward_qubit(int): the wire this edge points from
                backward_qubit(int): the wire this edge points to
                node(DAG.Node):  the node this edge points to
            """
            s_node, s_qubit = self.successors[forward_qubit]
            self.connect(forward_qubit, backward_qubit, node)
            node.connect(backward_qubit, s_qubit, s_node)

        def erase(self):
            """
            erase this node from the DAG. It will:
            1. set this node as FLAG_ERASED
            2. remove pointers between it and adjacent nodes
            """
            self.flag = self.FLAG_ERASED
            for qubit_ in range(self.size):
                p_node, p_qubit = self.predecessors[qubit_]
                n_node, n_qubit = self.successors[qubit_]
                if p_node:
                    p_node.connect(p_qubit, n_qubit, n_node)
                elif n_node:
                    n_node.predecessors[qubit_] = (None, 0)
            self.predecessors = None
            self.successors = None

    class VirtualNode(Node):
        """
        Virtual DAG node. When we want to temporarily insert a node into a DAG
        but not actually change the DAG structure, we can create a virtual node.
        """
        def __init__(self, gate_: BasicGate, predecessors, successors):
            """
            Args:
                gate_(BasicGate): the gate
                predecessors(List[Tuple[DAG.Node, int]]): predecessors of this node
                successors(List[Tuple[DAG.Node, int]]): successors of this node
            """
            super().__init__(gate_=gate_)
            self.predecessors = predecessors
            self.successors = successors

    def __init__(self, gates: Circuit, build_toffoli=True):
        """
        Args:
            gates(Circuit): Circuit represented by this DAG
        """

        self._width = gates.width()

        self.start_nodes = [self.Node(qubit_=i) for i in range(self._width)]
        self.end_nodes = [self.Node(qubit_=i) for i in range(self._width)]
        self.global_phase = 0
        self.has_symbolic_rz = False
        self.init_size = self._build_graph(gates, build_toffoli)
        self.build_toffoli = build_toffoli

    def width(self):
        """
        Get number of qubits of this circuit
        Returns:
            int: Number of qubits
        """
        return self._width

    @staticmethod
    def _build_ccz(gate_):
        """CCX decomposition described in Nam et.al."""
        cgate = CompositeGate()
        with cgate:
            CX & [1, 2]
            T_dagger & 2
            CX & [0, 2]
            T & 2
            CX & [1, 2]
            T_dagger & 2
            CX & [0, 2]
            CX & [0, 1]
            T_dagger & 1
            CX & [0, 1]
            T & 0
            T & 1
            T & 2

        args = gate_.cargs + gate_.targs
        if len(args) == gate_.controls + gate_.targets:
            cgate & args
        return cgate.gates

    def _build_graph(self, gates: Circuit, build_toffoli):
        node_cnt = 0
        cur_nodes = self.start_nodes.copy()

        var_cnt = 0
        for idx, gate_ in enumerate(gates.gates):
            # decouple ccx building with dag
            if build_toffoli and (gate_.type == GateType.ccx or gate_.type == GateType.ccz):
                self.has_symbolic_rz = True
                gate_list = self._build_ccz(gate_)
                if gate_.type == GateType.ccx:
                    gate_list = [H & gate_.targ] + gate_list + [H & gate_.targ]

                node_cnt += len(gate_list)

                # create a new phase variable
                var = SymbolicPhaseVariable(var_cnt)
                var_cnt += 1

                # represent phase of T/Tdg gates with this variable
                for each in gate_list:
                    node = self.Node(each)
                    if node.gate_type == GateType.t:
                        node.gate_type = GateType.rz
                        node.params = [SymbolicPhase() + var]
                        self.global_phase += var / 2
                    elif node.gate_type == GateType.tdg:
                        node.gate_type = GateType.rz
                        node.params = [SymbolicPhase() - var]
                        self.global_phase -= var / 2

                    for qubit_ in list(chain(each.cargs, each.targs)):
                        cur_nodes[qubit_].add_forward_edge(qubit_, node)
                        cur_nodes[qubit_] = node
            else:
                node_cnt += 1
                node = self.Node(gate_)
                for qubit_ in list(chain(gate_.cargs, gate_.targs)):
                    cur_nodes[qubit_].add_forward_edge(qubit_, node)
                    cur_nodes[qubit_] = node

        for qubit_ in range(self._width):
            cur_nodes[qubit_].add_forward_edge(qubit_, self.end_nodes[qubit_])

        return node_cnt

    def get_circuit(self):
        """
        Generate circuit net list from this DAG.

        Returns:
            Circuit: Circuit equivalent to this DAG
        """

        circ = Circuit(self._width)
        mapping = {(id(node), 0): qubit_ for qubit_, node in enumerate(self.start_nodes)}
        for node in self.topological_sort():
            for qubit_ in range(node.size):
                pred, qubit2 = node.predecessors[qubit_]
                mapping[(id(node), qubit_)] = mapping[(id(pred), qubit2)]

            node.get_gate() | circ([mapping[(id(node), qubit_)] for qubit_ in range(node.size)])

        if not np.isclose(float(self.global_phase), 0):
            GPhase(self.global_phase) | circ(0)
        return circ

    def topological_sort(self, include_dummy=False):
        """
        Iterate over nodes in this DAG in topological order (ignore start nodes)

        Returns:
            Iterator[DAG.Node]: nodes in topological order
        """

        edge_count = {}
        queue = deque(self.start_nodes)
        while len(queue) > 0:
            cur = queue.popleft()
            if cur.gate_type is not None or include_dummy:
                yield cur
            for nxt, _ in cur.successors:
                if nxt is None:
                    continue
                if id(nxt) not in edge_count:
                    edge_count[id(nxt)] = nxt.size
                edge_count[id(nxt)] -= 1
                if edge_count[id(nxt)] == 0:
                    queue.append(nxt)

    @staticmethod
    def topological_sort_sub_circuit(prev_node, succ_node):
        """
        Sort the sub circuit between prev_node and succ_node in topological order.

        Returns:
            Iterator[DAG.Node]: nodes in topological order
        """
        end_set = set()
        for node_, qubit_ in filter(lambda x: x is not None, succ_node):
            end_set.add(id(node_))

        edge_count = {}
        queue = deque()
        for each in prev_node:
            if each is None:
                continue
            node_, qubit_ = each
            nxt, nxt_q = node_.successors[qubit_]
            if id(nxt) in end_set:
                continue

            if id(nxt) not in edge_count:
                edge_count[id(nxt)] = nxt.size
            edge_count[id(nxt)] -= 1
            if edge_count[id(nxt)] == 0:
                queue.append(nxt)

        while len(queue) > 0:
            cur = queue.popleft()
            yield cur

            for nxt, nxt_q in cur.successors:
                if nxt is None or id(nxt) in end_set:
                    continue
                if id(nxt) not in edge_count:
                    edge_count[id(nxt)] = nxt.size
                edge_count[id(nxt)] -= 1
                if edge_count[id(nxt)] == 0:
                    queue.append(nxt)

    def reset_flag(self):
        """
        Set flag of all nodes to FLAG_DEFAULT
        """
        for node in self.start_nodes:
            node.flag = node.FLAG_DEFAULT
            node.qubit_flag = [node.FLAG_DEFAULT] * node.size
        for node in self.end_nodes:
            node.flag = node.FLAG_DEFAULT
            node.qubit_flag = [node.FLAG_DEFAULT] * node.size
        for node in self.topological_sort():
            node.flag = node.FLAG_DEFAULT
            node.qubit_flag = [node.FLAG_DEFAULT] * node.size

    def set_qubit_loc(self):
        """
        Update qubit_loc of all nodes.
        """

        mapping = {(id(node), 0): qubit_ for qubit_, node in enumerate(self.start_nodes)}
        for i, node in enumerate(self.start_nodes):
            node.qubit_loc[0] = i
            node.qubit_id = {i: 0}
        for i, node in enumerate(self.end_nodes):
            node.qubit_loc[0] = i
            node.qubit_id = {i: 0}

        for node in self.topological_sort():
            for qubit_ in range(node.size):
                pred, qubit2 = node.predecessors[qubit_]
                mapping[(id(node), qubit_)] = mapping[(id(pred), qubit2)]
                node.qubit_loc[qubit_] = mapping[(id(node), qubit_)]

            node.qubit_id = {qubit_: i for i, qubit_ in enumerate(node.qubit_loc)}

    def compare_circuit(self, other: Tuple[Node, int], anchor_qubit: int,
                        flag_enabled: bool = False, dummy_rz: bool = False):
        """
        Compare this circuit with another circuit.
        This circuit starts from the first gate on the `anchor_qubit`.
        The other starts from (gate, wire) defined by variable `other`.
        Param values (such as phases in Rz) ignored.

        Args:
            other(Tuple[Node, int]): start point of the other circuit
            anchor_qubit(int): start point of this circuit.
            flag_enabled(bool): Whether consider `flag` field. If true,
                nodes already with FLAG_VISITED will be
                skipped. Nodes in the matching will be set FLAG_VISITED.
            dummy_rz(bool): Enable dummy_rz feature: when there is a rz is `self`
            but not in `other`, create a virtual Rz(0) in `other` to match them.

        Returns:
            Dict[int, DAG.Node]: mapping from id(node of this circuit) to
                matched node in the other circuit. If not matched, return None.
        """
        t_node, t_qubit = self.start_nodes[anchor_qubit].successors[0]
        o_node, o_qubit = other[0], (t_qubit if other[1] == -1 else other[1])
        if o_node.gate_type is None:
            return None

        if o_node.gate_type != t_node.gate_type or (t_qubit != o_qubit) or (o_node.flag and flag_enabled):
            return None

        mapping = {id(t_node): o_node}
        queue = deque([(t_node, o_node)])
        while len(queue) > 0:
            u, v = queue.popleft()
            for neighbors in ['predecessors', 'successors']:
                for qubit_ in range(u.size):
                    u_nxt, u_qubit = getattr(u, neighbors)[qubit_]
                    if not u_nxt.gate_type:
                        continue

                    v_nxt, v_qubit = getattr(v, neighbors)[qubit_]
                    if id(u_nxt) in mapping:
                        if dummy_rz and isinstance(mapping[id(u_nxt)], DAG.VirtualNode):
                            continue
                        if id(mapping[id(u_nxt)]) != id(v_nxt) or u_qubit != v_qubit:
                            return None
                        continue

                    # v_nxt fails to match u_nxt
                    if not v_nxt.gate_type or (v_nxt.flag and flag_enabled) or \
                            u_nxt.gate_type != v_nxt.gate_type or u_qubit != v_qubit:
                        # if u_nxt is a rz, put a virtual node in v_nxt
                        if dummy_rz and u_nxt.gate_type == GateType.rz:
                            vnode = DAG.VirtualNode(
                                Rz(0),
                                [(v, qubit_)] if neighbors == 'successors' else [(v_nxt, v_qubit)],
                                [(v_nxt, v_qubit)] if neighbors == 'successors' else [(v, qubit_)]
                            )
                            mapping[id(u_nxt)] = vnode
                            queue.append((u_nxt, vnode))
                            continue
                        return None

                    mapping[id(u_nxt)] = v_nxt
                    queue.append((u_nxt, v_nxt))

        if flag_enabled:
            for each in mapping.values():
                each.flag = DAG.Node.FLAG_VISITED
        return mapping

    @staticmethod
    def replace_circuit(mapping: Dict[int, Tuple[Node, int]], replacement, erase_old=True):
        """
        Replace a part of this DAG with `replacement` defined by `mapping`.

        Args:
            mapping(Dict[int, Tuple[Node, int]]): mapping from the replacement to this circuit
            replacement: new sub circuit
            erase_old: whether erase old circuit
        """
        replacement: DAG
        erase_queue = deque()
        for qubit_ in range(replacement.width()):
            # first node on qubit_ in replacement circuit
            r_node, r_qubit = replacement.start_nodes[qubit_].successors[0]
            # node previous to the node corresponding to t_node in original circuit
            if id(replacement.start_nodes[qubit_]) not in mapping:
                continue
            p_node, p_qubit = mapping[id(replacement.start_nodes[qubit_])]
            if erase_old:
                erase_queue.append(p_node.successors[p_qubit][0])

            # place r_node after p_node
            p_node.connect(p_qubit, r_qubit, r_node)

        for qubit_ in range(replacement.width()):
            # last node on qubit_ in replacement circuit
            r_node, r_qubit = replacement.end_nodes[qubit_].predecessors[0]
            # node successive to the node corresponding to t_node in original circuit
            if id(replacement.end_nodes[qubit_]) not in mapping:
                continue
            s_node, s_qubit = mapping[id(replacement.end_nodes[qubit_])]
            d_node, d_qubit = s_node.predecessors[s_qubit]
            d_node.successors[d_qubit] = (None, 0)

            # place s_node after r_node
            r_node.connect(r_qubit, s_qubit, s_node)

        if erase_old:
            while len(erase_queue) > 0:
                cur = erase_queue.popleft()
                if cur is None or cur.flag == cur.FLAG_ERASED:
                    continue
                for qubit_ in range(cur.size):
                    erase_queue.append(cur.successors[qubit_][0])

                cur.flag = cur.FLAG_ERASED
                cur.predecessors = None
                cur.successors = None

    @staticmethod
    def _get_reachable_relation(node: Node, qubit_: int) -> Set[Tuple[Tuple[int, int], int]]:
        visited = set()
        queue = deque([(node, qubit_)])
        while len(queue) > 0:
            cur, cur_q = queue.popleft()
            if cur.gate_type is None:
                continue
            nxt, _ = cur.successors[cur_q]
            if id(nxt) not in visited:
                for nxt_q in range(nxt.size):
                    queue.append((nxt, nxt_q))
                visited.add(id(nxt))

        reachable = {((id(node), qubit_), o) for o in visited}
        return reachable

    def get_reachable_relation(self) -> Set[Tuple[Tuple[int, int], int]]:
        """
        Get reachable relation of nodes, namely whether (node A, wire) can reach the other node B.

        Returns:
            Set[Tuple[Tuple[int, int], int]]: Reachable relation ((id(A), wire), id(B))
        """
        reachable = set()
        for each in self.topological_sort(include_dummy=True):
            for qubit_ in range(each.size):
                reachable.update(self._get_reachable_relation(each, qubit_))
        return reachable

    @staticmethod
    def get_reachable_set(node: Node, qubit_: int, succ_node=None):
        """
        Get a set of reachable nodes starting from (node, qubit_)

        Args:
            node(DAG.Node): starting node
            qubit_(int): starting wire
            succ_node(List[Tuple[DAG.Node, int]]): right boundary

        Returns:
            Set[int]: id of reachable nodes
        """
        term_set = set()
        if succ_node is not None:
            for node_, qubit_ in filter(lambda x: x is not None, succ_node):
                term_set.add(id(node_))

        visited = set()
        queue = deque([(node, qubit_)])
        while len(queue) > 0:
            cur, cur_q = queue.popleft()
            if cur.gate_type is None or id(cur) in term_set:
                continue
            nxt, _ = cur.successors[cur_q]
            if id(nxt) not in visited:
                for nxt_q in range(nxt.size):
                    queue.append((nxt, nxt_q))
                visited.add(id(nxt))

        return visited

    @staticmethod
    def _get_sub_circuit_reachable_relation(node, qubit_, term_set):
        visited = set()
        queue = deque([(node, qubit_)])
        while len(queue) > 0:
            cur, cur_q = queue.popleft()
            if id(cur) in term_set:
                continue
            nxt, _ = cur.successors[cur_q]
            if id(nxt) not in visited:
                for nxt_q in range(nxt.size):
                    queue.append((nxt, nxt_q))
                visited.add(id(nxt))

        reachable = {((id(node), qubit_), o) for o in visited}
        return reachable

    @classmethod
    def get_sub_circuit_reachable_relation(cls, prev_node, succ_node):
        """
        Get reachable relation of nodes in the sub circuit,
        namely whether (node A, wire) can reach node B.

        If prev_node[i] and succ_node[i] is None, the sub circuit does not contain qubit i.

        Args:
            prev_node(List[Tuple[DAG.Node, int]]): left bound of the sub circuit.
            succ_node(List[Tuple[DAG.Node, int]]): right bound of the sub circuit.

        Returns:
            Set[Tuple[Tuple[int, int], int]]: Reachable relation ((id(A), wire), id(B))

        """
        term_set = set()
        for node_, qubit_ in filter(lambda x: x is not None, succ_node):
            term_set.add(id(node_))

        reachable = set()
        for each in cls.topological_sort_sub_circuit(prev_node, succ_node):
            for qubit_ in range(each.size):
                reachable.update(cls._get_sub_circuit_reachable_relation(each, qubit_, term_set))
        return reachable

    def append(self, gate_: BasicGate):
        """
        Add a gate after this DAG.

        Args:
            gate_(BasicGate): gate to add
        """
        node_ = self.Node(gate_)
        for wire_, qubit_ in enumerate(node_.qubit_loc):
            p_node, p_qubit = self.end_nodes[qubit_].predecessors[0]
            p_node.successors[p_qubit] = (node_, wire_)
            node_.predecessors[wire_] = (p_node, p_qubit)

            node_.successors[wire_] = (self.end_nodes[qubit_], 0)
            self.end_nodes[qubit_].predecessors[0] = (node_, wire_)

    def extend(self, gates: List[BasicGate]):
        """
        Add a list of gates after this DAG

        Args:
            gates(List[BasicGate]): gates to add
        """
        for each in gates:
            self.append(each)

    def __iter__(self):
        """
        Iterate over gates in this DAG in topological order

        Returns:
            Iterator[BasicGate]: gates in topological order
        """

        for node in self.topological_sort():
            yield node.get_gate()

    def copy(self):
        """
        Get a copy of this circuit

        Returns:
            DAG: a copy
        """
        return DAG(self.get_circuit(), build_toffoli=self.build_toffoli)

    def destroy(self):
        """
        Destroy this DAG.
        """
        que = deque(self.start_nodes)
        while len(que) > 0:
            cur = que.popleft()
            if cur is None or cur.flag == cur.FLAG_ERASED:
                continue
            for qubit_ in range(cur.size):
                que.append(cur.successors[qubit_][0])

            cur.flag = cur.FLAG_ERASED
            cur.predecessors = None
            cur.successors = None
